# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd
from orca import orca

from activitysim.core import (
    config,
    expressions,
    inject,
    logit,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.pathbuilder import TransitVirtualPathBuilder
from activitysim.core.util import assign_in_place, reindex

from .util import estimation, trip
from .util.mode import run_tour_mode_choice_simulate

logger = logging.getLogger(__name__)

"""
Tour mode choice is run for all tours to determine the transportation mode that
will be used for the tour
"""


def get_alts_from_segmented_nested_logit(model_settings, segment_name, trace_label):
    """Infer alts from logit spec

    Parameters
    ----------
    model_settings : dict
    segment_column_name : str
    trace_label : str

    Returns
    -------
    list
    """

    nest_spec = config.get_logit_model_settings(model_settings)
    coefficients = simulate.get_segment_coefficients(model_settings, segment_name)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients, trace_label)
    tour_mode_alts = []
    for nest in logit.each_nest(nest_spec):
        if nest.is_leaf:
            tour_mode_alts.append(nest.name)

    return tour_mode_alts


def create_logsum_trips(tours, segment_column_name, model_settings, trace_label):
    """
    Construct table of trips from half-tours (1 inbound, 1 outbound) for each tour-mode.

    Parameters
    ----------
    tours : pandas.DataFrame
    segment_column_name : str
        column in tours table used for segmenting model spec
    model_settings : dict
    trace_label : str

    Returns
    -------
    pandas.DataFrame
        Table of trips: 2 per tour, with O/D and purpose inherited from tour
    """
    stop_frequency_alts = inject.get_injectable("stop_frequency_alts")
    stop_freq = "0out_0in"  # no intermediate stops
    tours["stop_frequency"] = stop_freq
    tours["primary_purpose"] = tours["tour_purpose"]
    trips = trip.initialize_from_tours(tours, stop_frequency_alts)
    trips["stop_frequency"] = stop_freq
    outbound = trips["outbound"]
    trips["depart"] = reindex(tours.start, trips.tour_id)
    trips.loc[~outbound, "depart"] = reindex(tours.end, trips.loc[~outbound, "tour_id"])

    # actual segment doesn't matter. just need to grab one
    # to get a set of coefficients from the spec
    segment_name = tours.iloc[0][segment_column_name]
    tour_mode_alts = get_alts_from_segmented_nested_logit(
        model_settings, segment_name, trace_label
    )

    # repeat rows from the trips table iterating over tour mode
    logsum_trips = pd.DataFrame()
    for tour_mode in tour_mode_alts:
        trips["tour_mode"] = tour_mode
        logsum_trips = pd.concat((logsum_trips, trips), ignore_index=True)
    assert len(logsum_trips) == len(trips) * len(tour_mode_alts)
    logsum_trips.index.name = "trip_id"

    return logsum_trips


def append_tour_leg_trip_mode_choice_logsums(tours):
    """Creates trip mode choice logsum column in tours table for each tour mode and leg

    Parameters
    ----------
    tours : pd.DataFrame

    Returns
    -------
    tours : pd.DataFrame
        Adds two * n_modes logsum columns to each tour row, e.g. "logsum_DRIVE_outbound"
    """
    trips = inject.get_table("trips").to_frame()
    trip_dir_mode_logsums = trips.pivot(
        index="tour_id",
        columns=["tour_mode", "outbound"],
        values="trip_mode_choice_logsum",
    )
    new_cols = [
        "_".join(["logsum", mode, "outbound" if outbound else "inbound"])
        for mode, outbound in trip_dir_mode_logsums.columns
    ]
    trip_dir_mode_logsums.columns = new_cols
    trip_dir_mode_logsums.reindex(tours.index)
    tours = pd.merge(tours, trip_dir_mode_logsums, left_index=True, right_index=True)

    return tours


def get_trip_mc_logsums_for_all_modes(
    tours, segment_column_name, model_settings, trace_label
):
    """Creates pseudo-trips from tours and runs trip mode choice to get logsums

    Parameters
    ----------
    tours : pandas.DataFrame
    segment_column_name : str
        column in tours table used for segmenting model spec
    model_settings : dict
    trace_label : str

    Returns
    -------
    tours : pd.DataFrame
        Adds two * n_modes logsum columns to each tour row, e.g. "logsum_DRIVE_outbound"
    """

    # create pseudo-trips from tours for all tour modes
    logsum_trips = create_logsum_trips(
        tours, segment_column_name, model_settings, trace_label
    )

    # temporarily register trips in the pipeline
    pipeline.replace_table("trips", logsum_trips)
    tracing.register_traceable_table("trips", logsum_trips)
    pipeline.get_rn_generator().add_channel("trips", logsum_trips)

    # run trip mode choice on pseudo-trips. use orca instead of pipeline to
    # execute the step because pipeline can only handle one open step at a time
    orca.run(["trip_mode_choice"])

    # add trip mode choice logsums as new cols in tours
    tours = append_tour_leg_trip_mode_choice_logsums(tours)

    # de-register logsum trips table
    pipeline.get_rn_generator().drop_channel("trips")
    tracing.deregister_traceable_table("trips")

    return tours


@inject.step()
def tour_mode_choice_simulate(
    tours, persons_merged, network_los, chunk_size, trace_hh_id
):
    """
    Tour mode choice simulate
    """
    trace_label = "tour_mode_choice"
    model_settings_file_name = "tour_mode_choice.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get("MODE_CHOICE_LOGSUM_COLUMN_NAME")
    mode_column_name = "tour_mode"
    segment_column_name = "tour_purpose"

    primary_tours = tours.to_frame()
    assert not (primary_tours.tour_category == "atwork").any()

    logger.info("Running %s with %d tours" % (trace_label, primary_tours.shape[0]))

    tracing.print_summary("tour_types", primary_tours.tour_type, value_counts=True)

    persons_merged = persons_merged.to_frame()
    primary_tours_merged = pd.merge(
        primary_tours,
        persons_merged,
        left_on="person_id",
        right_index=True,
        how="left",
        suffixes=("", "_r"),
    )

    constants = {}
    # model_constants can appear in expressions
    constants.update(config.get_model_constants(model_settings))

    skim_dict = network_los.get_default_skim_dict()

    # setup skim keys
    orig_col_name = "home_zone_id"
    dest_col_name = "destination"

    out_time_col_name = "start"
    in_time_col_name = "end"
    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="out_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="in_period"
    )
    odr_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="in_period"
    )
    dor_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="out_period"
    )
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,  # dot return skims for e.g. TNC bridge return fare
        "dor_skims": dor_skim_stack_wrapper,  # odt return skims for e.g. TNC bridge return fare
        "od_skims": od_skim_stack_wrapper,
        "orig_col_name": orig_col_name,
        "dest_col_name": dest_col_name,
        "out_time_col_name": out_time_col_name,
        "in_time_col_name": in_time_col_name,
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?

        tvpb = network_los.tvpb

        tvpb_logsum_odt = tvpb.wrap_logsum(
            orig_key=orig_col_name,
            dest_key=dest_col_name,
            tod_key="out_period",
            segment_key="demographic_segment",
            cache_choices=True,
            trace_label=trace_label,
            tag="tvpb_logsum_odt",
        )
        tvpb_logsum_dot = tvpb.wrap_logsum(
            orig_key=dest_col_name,
            dest_key=orig_col_name,
            tod_key="in_period",
            segment_key="demographic_segment",
            cache_choices=True,
            trace_label=trace_label,
            tag="tvpb_logsum_dot",
        )

        skims.update(
            {"tvpb_logsum_odt": tvpb_logsum_odt, "tvpb_logsum_dot": tvpb_logsum_dot}
        )

        # TVPB constants can appear in expressions
        if model_settings.get("use_TVPB_constants", True):
            constants.update(
                network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
            )

    # don't create estimation data bundle if trip mode choice is being called
    # from another model step (i.e. tour mode choice logsum creation)
    if pipeline.get_rn_generator().step_name != "tour_mode_choice_simulate":
        estimator = None
    else:
        estimator = estimation.manager.begin_estimation("tour_mode_choice")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        estimator.write_coefficients_template(model_settings=model_settings)
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # (run_tour_mode_choice_simulate writes choosers post-annotation)

    # FIXME should normalize handling of tour_type and tour_purpose
    # mtctm1 school tour_type includes univ, which has different coefficients from elementary and HS
    # we should either add this column when tours created or add univ to tour_types
    not_university = (primary_tours_merged.tour_type != "school") | ~(
        primary_tours_merged.is_university.astype(bool)
        if "is_university" in primary_tours_merged.columns
        else False
    )
    primary_tours_merged["tour_purpose"] = primary_tours_merged.tour_type.where(
        not_university, "univ"
    )

    # if trip logsums are used, run trip mode choice and append the logsums
    if model_settings.get("COMPUTE_TRIP_MODE_CHOICE_LOGSUMS", False):
        primary_tours_merged = get_trip_mc_logsums_for_all_modes(
            primary_tours_merged, segment_column_name, model_settings, trace_label
        )

    choices_list = []
    for tour_purpose, tours_segment in primary_tours_merged.groupby(
        segment_column_name
    ):

        logger.info(
            "tour_mode_choice_simulate tour_type '%s' (%s tours)"
            % (
                tour_purpose,
                len(tours_segment.index),
            )
        )

        if network_los.zone_system == los.THREE_ZONE:
            tvpb_logsum_odt.extend_trace_label(tour_purpose)
            tvpb_logsum_dot.extend_trace_label(tour_purpose)

        # name index so tracing knows how to slice
        assert tours_segment.index.name == "tour_id"

        choices_df = run_tour_mode_choice_simulate(
            tours_segment,
            tour_purpose,
            model_settings,
            mode_column_name=mode_column_name,
            logsum_column_name=logsum_column_name,
            network_los=network_los,
            skims=skims,
            constants=constants,
            estimator=estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(trace_label, tour_purpose),
            trace_choice_name="tour_mode_choice",
        )

        tracing.print_summary(
            "tour_mode_choice_simulate %s choices_df" % tour_purpose,
            choices_df.tour_mode,
            value_counts=True,
        )

        choices_list.append(choices_df)

    choices_df = pd.concat(choices_list)

    # add cached tvpb_logsum tap choices for modes specified in tvpb_mode_path_types
    if network_los.zone_system == los.THREE_ZONE:

        tvpb_mode_path_types = model_settings.get("tvpb_mode_path_types")
        if tvpb_mode_path_types is not None:
            for mode, path_types in tvpb_mode_path_types.items():

                for direction, skim in zip(
                    ["od", "do"], [tvpb_logsum_odt, tvpb_logsum_dot]
                ):

                    path_type = path_types[direction]
                    skim_cache = skim.cache[path_type]

                    print(f"mode {mode} direction {direction} path_type {path_type}")

                    for c in skim_cache:

                        dest_col = f"{direction}_{c}"

                        if dest_col not in choices_df:
                            choices_df[dest_col] = (
                                np.nan
                                if pd.api.types.is_numeric_dtype(skim_cache[c])
                                else ""
                            )
                        choices_df[dest_col].where(
                            choices_df.tour_mode != mode, skim_cache[c], inplace=True
                        )

    if estimator:
        estimator.write_choices(choices_df.tour_mode)
        choices_df.tour_mode = estimator.get_survey_values(
            choices_df.tour_mode, "tours", "tour_mode"
        )
        estimator.write_override_choices(choices_df.tour_mode)
        estimator.end_estimation()

    tracing.print_summary(
        "tour_mode_choice_simulate all tour type choices",
        choices_df.tour_mode,
        value_counts=True,
    )

    # so we can trace with annotations
    assign_in_place(primary_tours, choices_df)

    # update tours table with mode choice (and optionally logsums)
    all_tours = tours.to_frame()
    assign_in_place(all_tours, choices_df)

    pipeline.replace_table("tours", all_tours)

    # - annotate tours table
    if model_settings.get("annotate_tours"):
        tours = inject.get_table("tours").to_frame()
        expressions.assign_columns(
            df=tours,
            model_settings=model_settings.get("annotate_tours"),
            trace_label=tracing.extend_trace_label(trace_label, "annotate_tours"),
        )
        pipeline.replace_table("tours", tours)

    if trace_hh_id:
        tracing.trace_df(
            primary_tours,
            label=tracing.extend_trace_label(trace_label, mode_column_name),
            slicer="tour_id",
            index_label="tour_id",
            warn_if_empty=True,
        )



# marin tvpb example data processing
# Ben Stabler, ben.stabler@rsginc.com, 09/17/20

import openmatrix as omx
import pandas as pd

# command to run the underdevelopment example
# python simulation.py -c configs_3_zone_marin -d data_3_marin -o output_3_marin

# data processing at c:\projects\activitysim\marin

# 1 - fix skim names, put time periods at end and make all names unique

time_periods = ["AM", "EA", "EV", "MD", "PM"]
for tp in time_periods:
    taz_file = omx.open_file("HWYSKM" + tp + "_taz.omx")
    taz_file_rename = omx.open_file("HWYSKM" + tp + "_taz_rename.omx", "w")
    for mat_name in taz_file.list_matrices():
        taz_file_rename[mat_name + "__" + tp] = taz_file[mat_name][:]
        print(mat_name + "__" + tp)
    taz_file.close()
    taz_file_rename.close()

for tp in time_periods:
    for skim_set in ["SET1", "SET2", "SET3"]:
        tap_file = omx.open_file("transit_skims_" + tp + "_" + skim_set + ".omx")
        tap_file_rename = omx.open_file(
            "transit_skims_" + tp + "_" + skim_set + "_rename.omx", "w"
        )
        for mat_name in tap_file.list_matrices():
            tap_file_rename[mat_name + "_" + skim_set + "__" + tp] = tap_file[mat_name][
                :
            ]
            print(mat_name + "_" + skim_set + "__" + tp)
        tap_file.close()
        tap_file_rename.close()

# 2 - nearby skims need headers

maz_tap_walk = pd.read_csv(
    "2015_test_2019_02_13_Part3/skims/ped_distance_maz_tap.txt", header=None
)
maz_maz_walk = pd.read_csv(
    "2015_test_2019_02_13_Part3/skims/ped_distance_maz_maz.txt", header=None
)
maz_maz_bike = pd.read_csv(
    "2015_test_2019_02_13_Part3/skims/bike_distance_maz_maz.txt", header=None
)

maz_tap_walk.columns = [
    "MAZ",
    "TAP",
    "TAP",
    "WALK_TRANSIT_GEN_COST",
    "WALK_TRANSIT_DIST",
]
maz_maz_walk.columns = ["OMAZ", "DMAZ", "DMAZ", "WALK_GEN_COST", "WALK_DIST"]
maz_maz_bike.columns = ["OMAZ", "DMAZ", "DMAZ", "BIKE_GEN_COST", "BIKE_DIST"]

maz_tap_walk["WALK_TRANSIT_DIST"] = maz_tap_walk["WALK_TRANSIT_DIST"] / 5280  # miles
maz_maz_walk["WALK_DIST"] = maz_maz_walk["WALK_DIST"] / 5280  # miles
maz_maz_bike["BIKE_DIST"] = maz_maz_bike["BIKE_DIST"] / 5280  # miles

maz_tap_walk[["MAZ", "TAP", "WALK_TRANSIT_DIST"]].to_csv(
    "maz_tap_walk.csv", index=False
)
maz_maz_walk[["OMAZ", "DMAZ", "WALK_DIST"]].to_csv("maz_maz_walk.csv", index=False)
maz_maz_bike[["OMAZ", "DMAZ", "BIKE_DIST"]].to_csv("maz_maz_bike.csv", index=False)

# 3 - maz data

mazs = pd.read_csv("2015_test_2019_02_13_Part2/landuse/maz_data_withDensity.csv")
pcost = pd.read_csv("2015_test_2019_02_13/ctramp_output/mgraParkingCost.csv")

mazs = pd.concat([mazs, pcost], axis=1)
mazs = mazs.fillna(0)

tazs = pd.read_csv("2015_test_2019_02_13_Part2/landuse/taz_data.csv")
tazs = tazs.set_index("TAZ", drop=False)

mazs["TERMINALTIME"] = tazs["TERMINALTIME"].loc[mazs["TAZ"]].tolist()

mazs["zone_id"] = mazs["MAZ"]
mazs["county_id"] = mazs["CountyID"]
mazs = mazs.set_index("zone_id", drop=False)

mazs.to_csv("maz_data_asim.csv", index=False)

# 4 - accessibility data

access = pd.read_csv("2015_test_2019_02_13/ctramp_output/accessibilities.csv")
access = access.drop([0])
access["zone_id"] = access["mgra"]
access = access.set_index("zone_id", drop=False)
access.to_csv("access.csv", index=False)

# 5 - maz to tap drive data

taz_tap_drive = pd.read_csv("2015_test_2019_02_13_Part3/skims/drive_maz_taz_tap.csv")

taz_tap_drive = taz_tap_drive.pivot_table(
    index=["FTAZ", "TTAP"], values=["DTIME", "DDIST", "WDIST"], fill_value=0
)

taz_tap_drive.columns = list(map("".join, taz_tap_drive.columns))
taz_tap_drive = taz_tap_drive.reset_index()
taz_tap_drive = taz_tap_drive.set_index("FTAZ")
taz_tap_drive["TAP"] = taz_tap_drive["TTAP"]

taz_tap_drive = pd.merge(
    mazs[["MAZ", "TAZ"]], taz_tap_drive, left_on=["TAZ"], right_on=["FTAZ"]
)
taz_tap_drive[["MAZ", "TAP", "DDIST", "DTIME", "WDIST"]].to_csv(
    "maz_taz_tap_drive.csv", index=False
)

# 6 - tours file, we just need work tours

itour = pd.read_csv("2015_test_2019_02_13/ctramp_output/indivTourData_3.csv")
work_tours = itour[itour["tour_purpose"] == "Work"]

work_tours["tour_id"] = range(1, len(work_tours) + 1)
work_tours["household_id"] = work_tours["hh_id"]
work_tours = work_tours.set_index("tour_id", drop=False)

work_tours["destination"] = work_tours["dest_mgra"]

work_tours["start"] = work_tours["start_period"]
work_tours["end"] = work_tours["end_period"]
work_tours["tour_type"] = "work"

work_tours.to_csv("work_tours.csv", index=False)

# 7 - households

households = pd.read_csv("2015_test_2019_02_13_Part2/popsyn/households.csv")
households["household_id"] = households["HHID"]
households["home_zone_id"] = households["MAZ"]
households = households.set_index("household_id", drop=False)

households.to_csv("households_asim.csv", index=False)

# 8 - persons

persons = pd.read_csv("2015_test_2019_02_13_Part2/popsyn/persons.csv")
persons["person_id"] = persons["PERID"]
persons["household_id"] = persons["HHID"]
persons = persons.set_index("person_id", drop=False)

persons_output = pd.read_csv("2015_test_2019_02_13/ctramp_output/personData_3.csv")
persons_output = persons_output.set_index("person_id", drop=False)
persons["type"] = persons_output["type"].loc[persons.index]
persons["value_of_time"] = persons_output["value_of_time"].loc[persons.index]
persons["is_university"] = persons["type"] == "University student"
persons["fp_choice"] = persons_output["fp_choice"]

persons.to_csv("persons_asim.csv", index=False)

# 9 - replace existing pipeline tables for restart for now

# run simple three zone example and get output pipeline and then replace tables before tour mode choice
pipeline = pd.io.pytables.HDFStore("pipeline.h5")
pipeline.keys()

pipeline["/accessibility/compute_accessibility"] = access  # index zone_id
pipeline["/households/joint_tour_frequency"] = households  # index household_id
pipeline["/persons/non_mandatory_tour_frequency"] = persons  # index person_id
pipeline["/land_use/initialize_landuse"] = mazs  # index zone_id
pipeline["/tours/non_mandatory_tour_scheduling"] = work_tours  # index tour_id

pipeline.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import inject, mem, pipeline, tracing
from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


def read_raw_persons(households):

    df = read_input_table("persons")

    if inject.get_injectable("households_sliced", False):
        # keep only persons in the sampled households
        df = df[df.household_id.isin(households.index)]

    return df


@inject.table()
def persons(households, trace_hh_id):

    df = read_raw_persons(households)

    logger.info("loaded persons %s" % (df.shape,))

    # replace table function with dataframe
    inject.add_table("persons", df)

    pipeline.get_rn_generator().add_channel("persons", df)

    tracing.register_traceable_table("persons", df)
    if trace_hh_id:
        tracing.trace_df(df, "raw.persons", warn_if_empty=True)

    logger.debug(f"{len(df.household_id.unique())} unique household_ids in persons")
    logger.debug(f"{len(households.index.unique())} unique household_ids in households")
    assert not households.index.duplicated().any()
    assert not df.index.duplicated().any()

    persons_without_households = ~df.household_id.isin(households.index)
    if persons_without_households.any():
        logger.error(
            f"{persons_without_households.sum()} persons out of {len(persons)} without households\n"
            f"{pd.Series({'person_id': persons_without_households.index.values})}"
        )
        raise RuntimeError(
            f"{persons_without_households.sum()} persons with bad household_id"
        )

    households_without_persons = (
        df.groupby("household_id").size().reindex(households.index).isnull()
    )
    if households_without_persons.any():
        logger.error(
            f"{households_without_persons.sum()} households out of {len(households.index)} without  persons\n"
            f"{pd.Series({'household_id': households_without_persons.index.values})}"
        )
        raise RuntimeError(
            f"{households_without_persons.sum()} households with no persons"
        )

    return df


# another common merge for persons
@inject.table()
def persons_merged(persons, households, land_use, accessibility):

    return inject.merge_tables(
        persons.name, tables=[persons, households, land_use, accessibility]
    )



import glob
import hashlib
import os
import shutil
import sys

import pkg_resources
import requests
import yaml

PACKAGE = "activitysim"
EXAMPLES_DIR = "examples"
MANIFEST = "example_manifest.yaml"


def _example_path(resource):
    resource = os.path.join(EXAMPLES_DIR, resource)
    path = pkg_resources.resource_filename(PACKAGE, resource)

    return path


def _load_manifest():
    with open(_example_path(MANIFEST), "r") as f:
        manifest = yaml.safe_load(f.read())

    assert manifest, f"error: could not load {MANIFEST}"
    return {example["name"]: example for example in manifest}


EXAMPLES = _load_manifest()


def add_create_args(parser):
    """Create command args"""
    create_group = parser.add_mutually_exclusive_group(required=True)
    create_group.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list available example directories and exit",
    )
    create_group.add_argument(
        "-e", "--example", type=str, metavar="PATH", help="example directory to copy"
    )

    parser.add_argument(
        "-d",
        "--destination",
        type=str,
        metavar="PATH",
        default=os.getcwd(),
        help="path to new project directory (default: %(default)s)",
    )


def create(args):
    """
    Create a new ActivitySim configuration from an existing template.

    Use the -l flag to view a list of example configurations, then
    copy to your own working directory. These new project files can
    be run with the 'run' command.
    """

    if args.list:

        list_examples()
        sys.exit(0)

    if args.example:

        get_example(args.example, args.destination)
        sys.exit(0)


def list_examples():
    print("*** Available examples ***\n")

    ret = []
    for example in list(EXAMPLES.values()):
        del example["include"]
        ret.append(example)
        print(yaml.dump(example))

    return ret


def get_example(example_name, destination, benchmarking=False):
    """
    Copy project data to user-specified directory.

    Examples and their data are described in a manifest
    YAML file. Each example contains at least a `name` and
    `include` field which is a list of files/folders to include
    in the copied example.



    Parameters
    ----------

    example_name: str, name of the example to copy.
        Options can be found via list_examples()
    destination: name of target directory to copy files to.
        If the target directory already exists, project files
        will be copied into a subdirectory with the same name
        as the example
    benchmarking: bool
    """
    if example_name not in EXAMPLES:
        sys.exit(f"error: could not find example '{example_name}'")

    if os.path.isdir(destination):
        dest_path = os.path.join(destination, example_name)
    else:
        dest_path = destination

    example = EXAMPLES[example_name]
    itemlist = example.get("include", [])
    if benchmarking:
        itemlist.extend(example.get("benchmarking", []))

    for item in itemlist:

        # split include string into source/destination paths
        items = item.split()
        assets = items[0]
        if len(items) == 3:
            target_path = os.path.join(dest_path, items[1])
            sha256 = items[-1]
        elif len(items) == 2:
            target_path = os.path.join(dest_path, items[-1])
            sha256 = None
        else:
            target_path = dest_path
            sha256 = None

        if assets.startswith("http"):
            download_asset(assets, target_path, sha256)

        else:
            for asset_path in glob.glob(_example_path(assets)):
                copy_asset(asset_path, target_path, dirs_exist_ok=True)

    print(f"copied! new project files are in {os.path.abspath(dest_path)}")

    instructions = example.get("instructions")
    if instructions:
        print(instructions)


def copy_asset(asset_path, target_path, dirs_exist_ok=False):

    print(f"copying {os.path.basename(asset_path)} ...")
    if os.path.isdir(asset_path):
        target_path = os.path.join(target_path, os.path.basename(asset_path))
        shutil.copytree(asset_path, target_path, dirs_exist_ok=dirs_exist_ok)

    else:
        target_dir = os.path.dirname(target_path)
        if target_dir:
            os.makedirs(target_dir, exist_ok=True)
        shutil.copy(asset_path, target_path)


def download_asset(url, target_path, sha256=None):
    os.makedirs(os.path.dirname(target_path), exist_ok=True)
    if url.endswith(".gz") and not target_path.endswith(".gz"):
        target_path_dl = target_path + ".gz"
    else:
        target_path_dl = target_path
    if sha256 and os.path.isfile(target_path):
        computed_sha256 = sha256_checksum(target_path)
        if sha256 == computed_sha256:
            print(f"not re-downloading existing {os.path.basename(target_path)} ...")
            return
        else:
            print(f"re-downloading existing {os.path.basename(target_path)} ...")
            print(f"   expected checksum {sha256}")
            print(f"   computed checksum {computed_sha256}")
    else:
        print(f"downloading {os.path.basename(target_path)} ...")
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(target_path_dl, "wb") as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)
    if target_path_dl != target_path:
        import gzip

        with gzip.open(target_path_dl, "rb") as f_in:
            with open(target_path, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(target_path_dl)
    computed_sha256 = sha256_checksum(target_path)
    if sha256 and sha256 != computed_sha256:
        raise ValueError(
            f"downloaded {os.path.basename(target_path)} has incorrect checksum\n"
            f"   expected checksum {sha256}\n"
            f"   computed checksum {computed_sha256}"
        )
    elif not sha256:
        print(f"   computed checksum {computed_sha256}")


def sha256_checksum(filename, block_size=65536):
    sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            sha256.update(block)
    return sha256.hexdigest()



import larch

from .cdap import *
from .data_maker import *
from .general import *
from .location_choice import *
from .mode_choice import *
from .nonmand_tour_freq import *
from .scheduling import *
from .simple_simulate import *
from .stop_frequency import *


def component_model(name, *args, **kwargs):
    if isinstance(name, str):
        m = globals().get(f"{name}_model")
        if m:
            return m(*args, **kwargs)
        raise KeyError(f"no known {name}_model")
    else:
        models = []
        all_data = []
        for n in name:
            model, *data = component_model(n, *args, **kwargs)
            models.append(model)
            all_data.extend(data)
        if all_data:
            return ModelGroup(models), all_data
        else:
            return ModelGroup(models)



# ActivitySim
# See full license in LICENSE.txt.

import logging
import os
import sys

import numpy as np
import pandas as pd
import yaml

from activitysim.abm.models.util import canonical_ids as cid
from activitysim.abm.models.util import tour_frequency as tf
from activitysim.core.util import reindex

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
logger.addHandler(ch)

CONSTANTS = {}

SURVEY_TOUR_ID = "survey_tour_id"
SURVEY_PARENT_TOUR_ID = "survey_parent_tour_id"
SURVEY_PARTICIPANT_ID = "survey_participant_id"
SURVEY_TRIP_ID = "survey_trip_id"
ASIM_TOUR_ID = "tour_id"
ASIM_PARENT_TOUR_ID = "parent_tour_id"
ASIM_TRIP_ID = "trip_id"

ASIM_PARTICIPANT_ID = "participant_id"

survey_tables = {
    "households": {"file_name": "survey_households.csv", "index": "household_id"},
    "persons": {"file_name": "survey_persons.csv", "index": "person_id"},
    "tours": {"file_name": "survey_tours.csv"},
    "joint_tour_participants": {"file_name": "survey_joint_tour_participants.csv"},
    "trips": {"file_name": "survey_trips.csv"},
}

outputs = {
    "households": "override_households.csv",
    "persons": "override_persons.csv",
    "tours": "override_tours.csv",
    "joint_tour_participants": "override_joint_tour_participants.csv",
    "trips": "override_trips.csv",
}

control_tables = {
    "households": {"file_name": "final_households.csv", "index": "household_id"},
    "persons": {"file_name": "final_persons.csv", "index": "person_id"},
    "tours": {"file_name": "final_tours.csv"},
    "joint_tour_participants": {"file_name": "final_joint_tour_participants.csv"},
    "trips": {"file_name": "final_trips.csv"},
}
apply_controls = True
skip_controls = not apply_controls


def mangle_ids(ids):
    return ids * 10


def unmangle_ids(ids):
    return ids // 10


def infer_cdap_activity(persons, tours, joint_tour_participants):

    mandatory_tour_types = ["work", "school"]
    non_mandatory_tour_types = [
        "escort",
        "shopping",
        "othmaint",
        "othdiscr",
        "eatout",
        "social",
    ]

    num_mandatory_tours = (
        tours[tours.tour_type.isin(mandatory_tour_types)]
        .groupby("person_id")
        .size()
        .reindex(persons.index)
        .fillna(0)
        .astype(np.int8)
    )

    num_non_mandatory_tours = (
        tours[tours.tour_type.isin(non_mandatory_tour_types)]
        .groupby("person_id")
        .size()
        .reindex(persons.index)
        .fillna(0)
        .astype(np.int8)
    )

    num_joint_tours = (
        joint_tour_participants.groupby("person_id")
        .size()
        .reindex(persons.index)
        .fillna(0)
        .astype(np.int8)
    )

    num_non_mandatory_tours += num_joint_tours

    cdap_activity = pd.Series("H", index=persons.index)
    cdap_activity = cdap_activity.where(num_mandatory_tours == 0, "M")
    cdap_activity = cdap_activity.where(
        (cdap_activity == "M") | (num_non_mandatory_tours == 0), "N"
    )

    return cdap_activity


def infer_mandatory_tour_frequency(persons, tours):

    num_work_tours = (
        tours[tours.tour_type == "work"]
        .groupby("person_id")
        .size()
        .reindex(persons.index)
        .fillna(0)
        .astype(np.int8)
    )

    num_school_tours = (
        tours[tours.tour_type == "school"]
        .groupby("person_id")
        .size()
        .reindex(persons.index)
        .fillna(0)
        .astype(np.int8)
    )

    mtf = {
        0: "",
        1: "work1",
        2: "work2",
        10: "school1",
        20: "school2",
        11: "work_and_school",
    }

    mandatory_tour_frequency = (num_work_tours + num_school_tours * 10).map(mtf)

    return mandatory_tour_frequency


def infer_non_mandatory_tour_frequency(configs_dir, persons, tours):
    def read_alts():
        # escort,shopping,othmaint,othdiscr,eatout,social
        # 0,0,0,0,0,0
        # 0,0,0,1,0,0, ...
        alts = pd.read_csv(
            os.path.join(configs_dir, "non_mandatory_tour_frequency_alternatives.csv"),
            comment="#",
        )
        alts = alts.astype(np.int8)  # - NARROW
        return alts

    tours = tours[tours.tour_category == "non_mandatory"]

    alts = read_alts()
    tour_types = list(alts.columns.values)

    # tour_frequency is index in alts table
    alts["alt_id"] = alts.index

    # actual tour counts (may exceed counts envisioned by alts)
    unconstrained_tour_counts = pd.DataFrame(index=persons.index)
    for tour_type in tour_types:
        unconstrained_tour_counts[tour_type] = (
            tours[tours.tour_type == tour_type]
            .groupby("person_id")
            .size()
            .reindex(persons.index)
            .fillna(0)
            .astype(np.int8)
        )

    # unextend tour counts
    # activitysim extend tours counts based on a probability table
    # counts can only be extended if original count is between 1 and 4
    # and tours can only be extended if their count is at the max possible
    max_tour_counts = alts[tour_types].max(axis=0)
    constrained_tour_counts = pd.DataFrame(index=persons.index)
    for tour_type in tour_types:
        constrained_tour_counts[tour_type] = unconstrained_tour_counts[tour_type].clip(
            upper=max_tour_counts[tour_type]
        )

    # persons whose tours were constrained who aren't eligible for extension becuase they have > 4 constrained tours
    has_constrained_tours = (unconstrained_tour_counts != constrained_tour_counts).any(
        axis=1
    )
    print("%s persons with constrained tours" % (has_constrained_tours.sum()))
    too_many_tours = has_constrained_tours & constrained_tour_counts.sum(axis=1) > 4
    if too_many_tours.any():
        print("%s persons with too many tours" % (too_many_tours.sum()))
        print(constrained_tour_counts[too_many_tours])
        # not sure what to do about this. Throw out some tours? let them through?
        print("not sure what to do about this. Throw out some tours? let them through?")
        assert False

    # determine alt id corresponding to constrained_tour_counts
    # need to do index waltz because pd.merge doesn't preserve index in this case
    alt_id = (
        pd.merge(
            constrained_tour_counts.reset_index(),
            alts,
            left_on=tour_types,
            right_on=tour_types,
            how="left",
        )
        .set_index(persons.index.name)
        .alt_id
    )

    # did we end up with any tour frequencies not in alts?
    if alt_id.isna().any():
        bad_tour_frequencies = alt_id.isna()
        logger.warning("WARNING Bad joint tour frequencies\n\n")
        logger.warning(
            "\nWARNING Bad non_mandatory tour frequencies: num_tours\n%s"
            % constrained_tour_counts[bad_tour_frequencies]
        )
        logger.warning(
            "\nWARNING Bad non_mandatory tour frequencies: num_tours\n%s"
            % tours[
                tours.person_id.isin(persons.index[bad_tour_frequencies])
            ].sort_values("person_id")
        )
        bug

    tf = unconstrained_tour_counts.rename(
        columns={tour_type: "_%s" % tour_type for tour_type in tour_types}
    )
    tf["non_mandatory_tour_frequency"] = alt_id
    return tf


def infer_joint_tour_frequency(configs_dir, households, tours):
    def read_alts():
        # right now this file just contains the start and end hour
        alts = pd.read_csv(
            os.path.join(configs_dir, "joint_tour_frequency_alternatives.csv"),
            comment="#",
            index_col="alt",
        )
        alts = alts.astype(np.int8)  # - NARROW
        return alts

    alts = read_alts()
    tour_types = list(alts.columns.values)

    assert len(alts.index[(alts == 0).all(axis=1)]) == 1  # should be one zero_tours alt
    zero_tours_alt = alts.index[(alts == 0).all(axis=1)].values[0]

    alts["joint_tour_frequency"] = alts.index
    joint_tours = tours[tours.tour_category == "joint"]

    num_tours = pd.DataFrame(index=households.index)
    for tour_type in tour_types:
        joint_tour_is_tour_type = joint_tours.tour_type == tour_type
        if joint_tour_is_tour_type.any():
            num_tours[tour_type] = (
                joint_tours[joint_tour_is_tour_type]
                .groupby("household_id")
                .size()
                .reindex(households.index)
                .fillna(0)
            )
        else:
            logger.warning(
                "WARNING infer_joint_tour_frequency - no tours of type '%s'" % tour_type
            )
            num_tours[tour_type] = 0
    num_tours = num_tours.fillna(0).astype(np.int64)

    # need to do index waltz because pd.merge doesn't preserve index in this case
    jtf = pd.merge(
        num_tours.reset_index(),
        alts,
        left_on=tour_types,
        right_on=tour_types,
        how="left",
    ).set_index(households.index.name)

    if jtf.joint_tour_frequency.isna().any():
        bad_tour_frequencies = jtf.joint_tour_frequency.isna()
        logger.warning("WARNING Bad joint tour frequencies\n\n")
        logger.warning(
            "\nWARNING Bad joint tour frequencies: num_tours\n%s"
            % num_tours[bad_tour_frequencies]
        )
        logger.warning(
            "\nWARNING Bad joint tour frequencies: num_tours\n%s"
            % joint_tours[
                joint_tours.household_id.isin(households.index[bad_tour_frequencies])
            ]
        )
        bug

    logger.info(
        "infer_joint_tour_frequency: %s households with joint tours",
        (jtf.joint_tour_frequency != zero_tours_alt).sum(),
    )

    return jtf.joint_tour_frequency


def infer_joint_tour_composition(persons, tours, joint_tour_participants):
    """
    assign joint_tours a 'composition' column ('adults', 'children', or 'mixed')
    depending on the composition of the joint_tour_participants
    """
    joint_tours = tours[tours.tour_category == "joint"].copy()

    joint_tour_participants = pd.merge(
        joint_tour_participants,
        persons,
        left_on="person_id",
        right_index=True,
        how="left",
    )

    # FIXME - computed by asim annotate persons - not needed if embeded in asim and called just-in-time
    if "adult" not in joint_tour_participants:
        joint_tour_participants["adult"] = joint_tour_participants.age >= 18

    tour_has_adults = (
        joint_tour_participants[joint_tour_participants.adult]
        .groupby(SURVEY_TOUR_ID)
        .size()
        .reindex(joint_tours[SURVEY_TOUR_ID])
        .fillna(0)
        > 0
    )

    tour_has_children = (
        joint_tour_participants[~joint_tour_participants.adult]
        .groupby([SURVEY_TOUR_ID])
        .size()
        .reindex(joint_tours[SURVEY_TOUR_ID])
        .fillna(0)
        > 0
    )

    assert (tour_has_adults | tour_has_children).all()

    joint_tours["composition"] = np.where(
        tour_has_adults, np.where(tour_has_children, "mixed", "adults"), "children"
    )

    return joint_tours.composition.reindex(tours.index).fillna("").astype(str)


def infer_tour_scheduling(configs_dir, tours):
    # given start and end periods, infer tdd

    def read_tdd_alts():
        # right now this file just contains the start and end hour
        tdd_alts = pd.read_csv(
            os.path.join(configs_dir, "tour_departure_and_duration_alternatives.csv")
        )
        tdd_alts["duration"] = tdd_alts.end - tdd_alts.start
        tdd_alts = tdd_alts.astype(np.int8)  # - NARROW

        tdd_alts["tdd"] = tdd_alts.index
        return tdd_alts

    tdd_alts = read_tdd_alts()

    if not tours.start.isin(tdd_alts.start).all():
        print(tours[~tours.start.isin(tdd_alts.start)])
    assert tours.start.isin(tdd_alts.start).all(), "not all tour starts in tdd_alts"

    assert tours.end.isin(tdd_alts.end).all(), "not all tour starts in tdd_alts"

    tdds = pd.merge(
        tours[["start", "end"]],
        tdd_alts,
        left_on=["start", "end"],
        right_on=["start", "end"],
        how="left",
    )

    if tdds.tdd.isna().any():
        bad_tdds = tours[tdds.tdd.isna()]
        print("Bad tour start/end times:")
        print(bad_tdds)
        bug

    # print("tdd_alts\n%s" %tdd_alts, "\n")
    # print("tours\n%s" %tours[['start', 'end']])
    # print("tdds\n%s" %tdds)
    return tdds.tdd


def patch_tour_ids(persons, tours, joint_tour_participants):
    def set_tour_index(tours, parent_tour_num_col, is_joint):

        group_cols = ["person_id", "tour_category", "tour_type"]

        if "parent_tour_num" in tours:
            group_cols += ["parent_tour_num"]

        tours["tour_type_num"] = (
            tours.sort_values(by=group_cols).groupby(group_cols).cumcount() + 1
        )

        return cid.set_tour_index(
            tours, parent_tour_num_col=parent_tour_num_col, is_joint=is_joint
        )

    assert "mandatory_tour_frequency" in persons

    # replace survey_tour ids with asim standard tour_ids (which are based on person_id and tour_type)

    #####################
    # mandatory tours
    #####################
    mandatory_tours = set_tour_index(
        tours[tours.tour_category == "mandatory"],
        parent_tour_num_col=None,
        is_joint=False,
    )

    assert mandatory_tours.index.name == "tour_id"

    #####################
    # joint tours
    #####################

    # joint tours tour_id was assigned based on person_id of the first person in household (PNUM == 1)
    # because the actual point person forthe tour is only identified later in joint_tour_participants)
    temp_point_persons = persons.loc[persons.PNUM == 1, ["household_id"]]
    temp_point_persons["person_id"] = temp_point_persons.index
    temp_point_persons.set_index("household_id", inplace=True)

    # patch person_id with value of temp_point_person_id and use it to set_tour_index
    joint_tours = tours[tours.tour_category == "joint"]
    joint_tours["cache_point_person_id"] = joint_tours["person_id"]
    joint_tours["person_id"] = reindex(
        temp_point_persons.person_id, joint_tours.household_id
    )

    joint_tours = set_tour_index(joint_tours, parent_tour_num_col=None, is_joint=True)
    joint_tours["person_id"] = joint_tours["cache_point_person_id"]
    del joint_tours["cache_point_person_id"]

    # patch tour_id column in patched_joint_tour_participants
    patched_joint_tour_participants = joint_tour_participants.copy()
    asim_tour_id = pd.Series(joint_tours.index, index=joint_tours[SURVEY_TOUR_ID])
    patched_joint_tour_participants[ASIM_TOUR_ID] = reindex(
        asim_tour_id, patched_joint_tour_participants[SURVEY_TOUR_ID]
    )

    # participant_id is formed by combining tour_id and participant pern.PNUM
    # pathological knowledge, but awkward to conflate with joint_tour_participation.py logic
    participant_pnum = reindex(
        persons.PNUM, patched_joint_tour_participants["person_id"]
    )
    patched_joint_tour_participants[ASIM_PARTICIPANT_ID] = (
        patched_joint_tour_participants[ASIM_TOUR_ID] * cid.MAX_PARTICIPANT_PNUM
    ) + participant_pnum

    #####################
    # non_mandatory tours
    #####################

    non_mandatory_tours = set_tour_index(
        tours[tours.tour_category == "non_mandatory"],
        parent_tour_num_col=None,
        is_joint=False,
    )

    #####################
    # atwork tours
    #####################

    atwork_tours = tours[tours.tour_category == "atwork"]

    # patch atwork tours parent_tour_id before assigning their tour_id

    # tours for workers with both work and school trips should have lower tour_num for work,
    # tours for students with both work and school trips should have lower tour_num for school
    # tours are already sorted, but schools comes before work (which is alphabetical, not the alternative id order),
    # so work_and_school tour_nums are correct for students (school=1, work=2) but workers need to be flipped
    mandatory_tour_frequency = reindex(
        persons.mandatory_tour_frequency, mandatory_tours.person_id
    )
    is_worker = reindex(persons.pemploy, mandatory_tours.person_id).isin(
        [CONSTANTS["PEMPLOY_FULL"], CONSTANTS["PEMPLOY_PART"]]
    )
    work_and_school_and_worker = (
        mandatory_tour_frequency == "work_and_school"
    ) & is_worker

    # calculate tour_num for work tours (required to set_tour_index for atwork subtours)

    parent_tours = mandatory_tours[[SURVEY_TOUR_ID]]
    parent_tours["tour_num"] = (
        mandatory_tours.sort_values(by=["person_id", "tour_category", "tour_type"])
        .groupby(["person_id", "tour_category"])
        .cumcount()
        + 1
    )

    parent_tours.tour_num = parent_tours.tour_num.where(
        ~work_and_school_and_worker, 3 - parent_tours.tour_num
    )
    parent_tours = parent_tours.set_index(SURVEY_TOUR_ID, drop=True)

    # temporarily add parent_tour_num column to atwork tours, call set_tour_index, and then delete it
    atwork_tours["parent_tour_num"] = reindex(
        parent_tours.tour_num, atwork_tours[SURVEY_PARENT_TOUR_ID]
    )

    atwork_tours = set_tour_index(
        atwork_tours, parent_tour_num_col="parent_tour_num", is_joint=False
    )

    del atwork_tours["parent_tour_num"]

    # tours['household_id'] = reindex(persons.household_id, tours.person_id)
    asim_tour_id = pd.Series(
        mandatory_tours.index, index=mandatory_tours[SURVEY_TOUR_ID]
    )
    atwork_tours[ASIM_PARENT_TOUR_ID] = reindex(
        asim_tour_id, atwork_tours[SURVEY_PARENT_TOUR_ID]
    )

    #####################
    # concat tours
    #####################

    # only true for fake data
    assert (
        mandatory_tours.index == unmangle_ids(mandatory_tours[SURVEY_TOUR_ID])
    ).all()
    assert (joint_tours.index == unmangle_ids(joint_tours[SURVEY_TOUR_ID])).all()
    assert (
        non_mandatory_tours.index == unmangle_ids(non_mandatory_tours[SURVEY_TOUR_ID])
    ).all()

    patched_tours = pd.concat(
        [mandatory_tours, joint_tours, non_mandatory_tours, atwork_tours]
    )

    assert patched_tours.index.name == ASIM_TOUR_ID
    patched_tours = patched_tours.reset_index()

    del patched_tours["tour_type_num"]

    assert ASIM_TOUR_ID in patched_tours
    assert ASIM_PARENT_TOUR_ID in patched_tours

    return patched_tours, patched_joint_tour_participants


def infer_atwork_subtour_frequency(configs_dir, tours):

    # first column is 'atwork_subtour_frequency' nickname, remaining columns are trip type counts
    alts = pd.read_csv(
        os.path.join(configs_dir, "atwork_subtour_frequency_alternatives.csv"),
        comment="#",
    )
    tour_types = list(
        alts.drop(columns=alts.columns[0]).columns
    )  # get trip_types, ignoring first column
    alts["alt_id"] = alts.index

    #             alt  eat  business  maint  alt_id
    # 0   no_subtours    0         0      0       0
    # 1           eat    1         0      0       1
    # 2     business1    0         1      0       2
    # 3         maint    0         0      1       3
    # 4     business2    0         2      0       4
    # 5  eat_business    1         1      0       5

    work_tours = tours[tours.tour_type == "work"]
    work_tours = work_tours[[ASIM_TOUR_ID]]

    subtours = tours[tours.tour_category == "atwork"]
    subtours = subtours[["tour_id", "tour_type", "parent_tour_id"]]

    # actual tour counts (may exceed counts envisioned by alts)
    tour_counts = pd.DataFrame(index=work_tours[ASIM_TOUR_ID])
    for tour_type in tour_types:
        # count subtours of this type by parent_tour_id
        tour_type_count = (
            subtours[subtours.tour_type == tour_type].groupby("parent_tour_id").size()
        )
        # backfill with 0 count
        tour_counts[tour_type] = (
            tour_type_count.reindex(tour_counts.index).fillna(0).astype(np.int8)
        )

    # determine alt id corresponding to constrained_tour_counts
    # need to do index waltz because pd.merge doesn't preserve index in this case
    tour_counts = pd.merge(
        tour_counts.reset_index(),
        alts,
        left_on=tour_types,
        right_on=tour_types,
        how="left",
    ).set_index(tour_counts.index.name)

    atwork_subtour_frequency = tour_counts.alt

    # did we end up with any tour frequencies not in alts?
    if atwork_subtour_frequency.isna().any():
        bad_tour_frequencies = atwork_subtour_frequency.isna()
        logger.warning(
            "WARNING Bad atwork subtour frequencies for %s work tours"
            % bad_tour_frequencies.sum()
        )
        logger.warning(
            "WARNING Bad atwork subtour frequencies: num_tours\n%s"
            % tour_counts[bad_tour_frequencies]
        )
        logger.warning(
            "WARNING Bad atwork subtour frequencies: num_tours\n%s"
            % subtours[
                subtours.parent_tour_id.isin(tour_counts[bad_tour_frequencies].index)
            ].sort_values("parent_tour_id")
        )
        bug

    atwork_subtour_frequency = reindex(
        atwork_subtour_frequency, tours[ASIM_TOUR_ID]
    ).fillna("")

    return atwork_subtour_frequency


def patch_trip_ids(tours, trips):
    """
    replace survey trip_ids with asim standard trip_id
    replace survey tour_id foreign key with asim standard tour_id
    """

    # tour_id is a column, not index
    assert ASIM_TOUR_ID in tours

    # patch tour_id foreign key
    # tours['household_id'] = reindex(persons.household_id, tours.person_id)
    asim_tour_id = pd.Series(
        tours[ASIM_TOUR_ID].values, index=tours[SURVEY_TOUR_ID].values
    )
    trips[ASIM_TOUR_ID] = reindex(asim_tour_id, trips[SURVEY_TOUR_ID])

    # person_is_university = persons.pstudent == constants.PSTUDENT_UNIVERSITY
    # tour_is_university = reindex(person_is_university, tours.person_id)
    # tour_primary_purpose = tours.tour_type.where((tours.tour_type != 'school') | ~tour_is_university, 'univ')
    # tour_primary_purpose = tour_primary_purpose.where(tours.tour_category!='atwork', 'atwork')
    #
    # trips['primary_purpose'] = reindex(tour_primary_purpose, trips.tour_id)

    # if order is ambiguous if trips depart in same time slot - order by SURVEY_TRIP_ID hoping that increases with time
    if "trip_num" not in trips:
        trips["trip_num"] = (
            trips.sort_values(by=["tour_id", "outbound", "depart", SURVEY_TRIP_ID])
            .groupby(["tour_id", "outbound"])
            .cumcount()
            + 1
        )

    cid.set_trip_index(trips)

    assert trips.index.name == ASIM_TRIP_ID
    trips = trips.reset_index().rename(columns={"trip_id": ASIM_TRIP_ID})

    return trips


def infer_stop_frequency(configs_dir, tours, trips):

    # alt,out,in
    # 0out_0in,0,0
    # 0out_1in,0,1
    # ...
    alts = pd.read_csv(
        os.path.join(configs_dir, "stop_frequency_alternatives.csv"), comment="#"
    )
    assert "alt" in alts
    assert "in" in alts
    assert "out" in alts

    freq = pd.DataFrame(index=tours[SURVEY_TOUR_ID])

    # number of trips is one less than number of stops
    freq["out"] = trips[trips.outbound].groupby(SURVEY_TOUR_ID).trip_num.max() - 1
    freq["in"] = trips[~trips.outbound].groupby(SURVEY_TOUR_ID).trip_num.max() - 1

    freq = pd.merge(freq.reset_index(), alts, on=["out", "in"], how="left")

    assert (freq[SURVEY_TOUR_ID] == tours[SURVEY_TOUR_ID]).all()

    return freq.alt


def read_tables(input_dir, tables):

    for table, info in tables.items():
        table = pd.read_csv(
            os.path.join(input_dir, info["file_name"]), index_col=info.get("index")
        )
        # coerce missing data in string columns to empty strings, not NaNs
        for c in table.columns:
            # read_csv converts empty string to NaN, even if all non-empty values are strings
            if table[c].dtype == "object":
                print("##### converting", c, table[c].dtype)
                table[c] = table[c].fillna("").astype(str)
        info["table"] = table

    households = tables["households"].get("table")
    persons = tables["persons"].get("table")
    tours = tables["tours"].get("table")
    joint_tour_participants = tables["joint_tour_participants"].get("table")
    trips = tables["trips"].get("table")

    return households, persons, tours, joint_tour_participants, trips


def check_controls(table_name, column_name):

    table = survey_tables[table_name].get("table")
    c_table = control_tables[table_name].get("table")

    if column_name == "index":
        dont_match = table.index != c_table.index
    else:
        dont_match = table[column_name] != c_table[column_name]

    if dont_match.any():
        print(
            "check_controls %s.%s: %s out of %s do not match"
            % (table_name, column_name, dont_match.sum(), len(table))
        )
        print("control\n%s" % c_table[dont_match][[column_name]])
        print("survey\n%s" % table[dont_match][[column_name]])

        print("control\n%s" % c_table[dont_match][table.columns])
        print("survey\n%s" % table[dont_match][table.columns])
        return False

    return True


def infer(configs_dir, input_dir, output_dir):

    households, persons, tours, joint_tour_participants, trips = read_tables(
        input_dir, survey_tables
    )

    # be explicit about all tour_ids to avoid confusion between asim and survey ids
    tours = tours.rename(
        columns={"tour_id": SURVEY_TOUR_ID, "parent_tour_id": SURVEY_PARENT_TOUR_ID}
    )
    joint_tour_participants = joint_tour_participants.rename(
        columns={"tour_id": SURVEY_TOUR_ID, "participant_id": SURVEY_PARTICIPANT_ID}
    )
    trips = trips.rename(columns={"trip_id": SURVEY_TRIP_ID, "tour_id": SURVEY_TOUR_ID})

    # mangle survey tour ids to keep us honest
    tours[SURVEY_TOUR_ID] = mangle_ids(tours[SURVEY_TOUR_ID])
    tours[SURVEY_PARENT_TOUR_ID] = mangle_ids(tours[SURVEY_PARENT_TOUR_ID])
    joint_tour_participants[SURVEY_TOUR_ID] = mangle_ids(
        joint_tour_participants[SURVEY_TOUR_ID]
    )
    joint_tour_participants[SURVEY_PARTICIPANT_ID] = mangle_ids(
        joint_tour_participants[SURVEY_PARTICIPANT_ID]
    )
    trips[SURVEY_TRIP_ID] = mangle_ids(trips[SURVEY_TRIP_ID])
    trips[SURVEY_TOUR_ID] = mangle_ids(trips[SURVEY_TOUR_ID])

    # persons.cdap_activity
    persons["cdap_activity"] = infer_cdap_activity(
        persons, tours, joint_tour_participants
    )
    # check but don't assert as this is not deterministic
    skip_controls or check_controls("persons", "cdap_activity")

    # persons.mandatory_tour_frequency
    persons["mandatory_tour_frequency"] = infer_mandatory_tour_frequency(persons, tours)
    assert skip_controls or check_controls("persons", "mandatory_tour_frequency")

    # persons.non_mandatory_tour_frequency
    tour_frequency = infer_non_mandatory_tour_frequency(configs_dir, persons, tours)
    for c in tour_frequency.columns:
        print("assigning persons", c)
        persons[c] = tour_frequency[c]
    assert skip_controls or check_controls("persons", "non_mandatory_tour_frequency")

    # patch_tour_ids
    tours, joint_tour_participants = patch_tour_ids(
        persons, tours, joint_tour_participants
    )
    survey_tables["tours"]["table"] = tours
    survey_tables["joint_tour_participants"]["table"] = joint_tour_participants

    assert skip_controls or check_controls("tours", "index")
    assert skip_controls or check_controls("joint_tour_participants", "index")

    # patch_tour_ids
    trips = patch_trip_ids(tours, trips)
    survey_tables["trips"]["table"] = trips  # so we can check_controls
    assert skip_controls or check_controls("trips", "index")

    # households.joint_tour_frequency
    households["joint_tour_frequency"] = infer_joint_tour_frequency(
        configs_dir, households, tours
    )
    assert skip_controls or check_controls("households", "joint_tour_frequency")

    # tours.composition
    tours["composition"] = infer_joint_tour_composition(
        persons, tours, joint_tour_participants
    )
    assert skip_controls or check_controls("tours", "composition")

    # tours.tdd
    tours["tdd"] = infer_tour_scheduling(configs_dir, tours)
    assert skip_controls or check_controls("tours", "tdd")

    tours["atwork_subtour_frequency"] = infer_atwork_subtour_frequency(
        configs_dir, tours
    )
    assert skip_controls or check_controls("tours", "atwork_subtour_frequency")

    tours["stop_frequency"] = infer_stop_frequency(configs_dir, tours, trips)
    assert skip_controls or check_controls("tours", "stop_frequency")

    # write output files
    households.to_csv(os.path.join(output_dir, outputs["households"]), index=True)
    persons.to_csv(os.path.join(output_dir, outputs["persons"]), index=True)
    tours.to_csv(os.path.join(output_dir, outputs["tours"]), index=False)
    joint_tour_participants.to_csv(
        os.path.join(output_dir, outputs["joint_tour_participants"]), index=False
    )
    trips.to_csv(os.path.join(output_dir, outputs["trips"]), index=False)


# python infer.py data
args = sys.argv[1:]
assert len(args) == 2, "usage: python infer.py <data_dir> <configs_dir>"

data_dir = args[0]
configs_dir = args[1]

with open(os.path.join(configs_dir, "constants.yaml")) as stream:
    CONSTANTS = yaml.load(stream, Loader=yaml.SafeLoader)

input_dir = os.path.join(data_dir, "survey_data/")
output_dir = input_dir

if apply_controls:
    read_tables(input_dir, control_tables)

infer(configs_dir, input_dir, output_dir)



import os

import orca
import pandas as pd
import pytest

from activitysim.core import pipeline
from activitysim.core.los import Network_LOS as los


@pytest.fixture(scope="module")
def initialize_pipeline(
    module: str, tables: dict[str, str], initialize_network_los: bool
) -> pipeline.Pipeline:
    test_dir = os.path.join("test", module)
    configs_dir = os.path.join(test_dir, "configs")
    data_dir = os.path.join(test_dir, "data")
    output_dir = os.path.join(test_dir, "output")

    if os.path.isdir(configs_dir):
        orca.add_injectable("configs_dir", configs_dir)

    if os.path.isdir(data_dir):
        orca.add_injectable("data_dir", data_dir)

    if os.path.isdir(test_dir):
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)
        orca.add_injectable("output_dir", output_dir)

    # Read in the input test dataframes
    for dataframe_name, idx_name in tables.items():
        df = pd.read_csv(
            os.path.join("test", module, "data", f"{dataframe_name}.csv"),
            index_col=idx_name,
        )
        orca.add_table(dataframe_name, df)

    if initialize_network_los:
        net_los = los()
        net_los.load_data()
        orca.add_injectable("network_los", net_los)

    # Add the dataframes to the pipeline
    pipeline.open_pipeline()
    pipeline.add_checkpoint(module)
    pipeline.close_pipeline()

    # By convention, this method needs to yield something
    yield pipeline._PIPELINE

    # pytest teardown code
    pipeline.close_pipeline()
    pipeline_file_path = os.path.join(output_dir, "pipeline.h5")
    os.unlink(pipeline_file_path)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import inject
from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


@inject.table()
def accessibility(land_use):
    """
    If 'accessibility' is in input_tables list, then read it in,
    otherwise create skeleton table with same index as landuse.

    This allows loading of pre-computed accessibility table, which is particularly useful
    for single-process small household sample runs when there are many zones in landuse

    skeleton table only required if multiprocessing wants to slice accessibility,
    otherwise it will simply be replaced when accessibility model is run
    """

    accessibility_df = read_input_table("accessibility", required=False)

    if accessibility_df is None:
        accessibility_df = pd.DataFrame(index=land_use.index)
        logger.debug(
            "created placeholder accessibility table %s" % (accessibility_df.shape,)
        )
    else:
        assert accessibility_df.sort_index().index.equals(
            land_use.to_frame().sort_index().index
        ), f"loaded accessibility table index does not match index of land_use table"
        logger.info("loaded land_use %s" % (accessibility_df.shape,))

    # replace table function with dataframe
    inject.add_table("accessibility", accessibility_df)

    return accessibility_df



# ActivitySim
# See full license in LICENSE.txt.

import logging
import warnings
from builtins import range
from collections import OrderedDict

import numpy as np
import pandas as pd

from . import assign, chunk, config, logit, pathbuilder, pipeline, tracing, util

logger = logging.getLogger(__name__)

SPEC_DESCRIPTION_NAME = "Description"
SPEC_EXPRESSION_NAME = "Expression"
SPEC_LABEL_NAME = "Label"

ALT_LOSER_UTIL = -900


def random_rows(df, n):

    # only sample if df has more than n rows
    if len(df.index) > n:
        prng = pipeline.get_rn_generator().get_global_rng()
        return df.take(prng.choice(len(df), size=n, replace=False))

    else:
        return df


def uniquify_spec_index(spec):

    # uniquify spec index inplace
    # ensure uniqueness of spec index by appending comment with dupe count
    # this allows us to use pandas dot to compute_utilities
    dict = OrderedDict()
    for expr in spec.index:
        dict[assign.uniquify_key(dict, expr, template="{} # ({})")] = expr

    prev_index_name = spec.index.name
    spec.index = list(dict.keys())
    spec.index.name = prev_index_name

    assert spec.index.is_unique


def read_model_alts(file_name, set_index=None):
    file_path = config.config_file_path(file_name)
    df = pd.read_csv(file_path, comment="#")
    if set_index:
        df.set_index(set_index, inplace=True)
    return df


def read_model_spec(file_name):
    """
    Read a CSV model specification into a Pandas DataFrame or Series.

    file_path : str   absolute or relative path to file

    The CSV is expected to have columns for component descriptions
    and expressions, plus one or more alternatives.

    The CSV is required to have a header with column names. For example:

        Description,Expression,alt0,alt1,alt2

    Parameters
    ----------
    model_settings : dict
        name of spec_file is in model_settings['SPEC'] and file is relative to configs
    file_name : str
        file_name id spec file in configs folder

    description_name : str, optional
        Name of the column in `fname` that contains the component description.
    expression_name : str, optional
        Name of the column in `fname` that contains the component expression.

    Returns
    -------
    spec : pandas.DataFrame
        The description column is dropped from the returned data and the
        expression values are set as the table index.
    """

    assert isinstance(file_name, str)
    if not file_name.lower().endswith(".csv"):
        file_name = "%s.csv" % (file_name,)

    file_path = config.config_file_path(file_name)

    try:
        spec = pd.read_csv(file_path, comment="#")
    except Exception as err:
        logger.error(f"read_model_spec error reading {file_path}")
        logger.error(f"read_model_spec error {type(err).__name__}: {str(err)}")
        raise (err)

    spec = spec.dropna(subset=[SPEC_EXPRESSION_NAME])

    # don't need description and set the expression to the index
    if SPEC_DESCRIPTION_NAME in spec.columns:
        spec = spec.drop(SPEC_DESCRIPTION_NAME, axis=1)

    spec = spec.set_index(SPEC_EXPRESSION_NAME).fillna(0)

    # ensure uniqueness of spec index by appending comment with dupe count
    # this allows us to use pandas dot to compute_utilities
    uniquify_spec_index(spec)

    if SPEC_LABEL_NAME in spec:
        spec = spec.set_index(SPEC_LABEL_NAME, append=True)
        assert isinstance(spec.index, pd.MultiIndex)

    return spec


def read_model_coefficients(model_settings=None, file_name=None):
    """
    Read the coefficient file specified by COEFFICIENTS model setting
    """

    if model_settings is None:
        assert file_name is not None
    else:
        assert file_name is None
        assert (
            "COEFFICIENTS" in model_settings
        ), "'COEFFICIENTS' tag not in model_settings in %s" % model_settings.get(
            "source_file_paths"
        )
        file_name = model_settings["COEFFICIENTS"]
        logger.debug(f"read_model_coefficients file_name {file_name}")

    file_path = config.config_file_path(file_name)
    try:
        coefficients = pd.read_csv(file_path, comment="#", index_col="coefficient_name")
    except ValueError:
        logger.exception("Coefficient File Invalid: %s" % str(file_path))
        raise

    if coefficients.index.duplicated().any():
        logger.warning(
            f"duplicate coefficients in {file_path}\n"
            f"{coefficients[coefficients.index.duplicated(keep=False)]}"
        )
        raise RuntimeError(f"duplicate coefficients in {file_path}")

    if coefficients.value.isnull().any():
        logger.warning(
            f"null coefficients in {file_path}\n{coefficients[coefficients.value.isnull()]}"
        )
        raise RuntimeError(f"null coefficients in {file_path}")

    return coefficients


def spec_for_segment(model_settings, spec_id, segment_name, estimator):
    """
    Select spec for specified segment from omnibus spec containing columns for each segment

    Parameters
    ----------
    model_spec : pandas.DataFrame
        omnibus spec file with expressions in index and one column per segment
    segment_name : str
        segment_name that is also column name in model_spec

    Returns
    -------
    pandas.dataframe
        canonical spec file with expressions in index and single column with utility coefficients
    """

    spec_file_name = model_settings[spec_id]
    spec = read_model_spec(file_name=spec_file_name)

    if len(spec.columns) > 1:
        # if spec is segmented
        spec = spec[[segment_name]]
    else:
        # otherwise we expect a single coefficient column
        # doesn't really matter what it is called, but this may catch errors
        assert spec.columns[0] in ["coefficient", segment_name]

    if "COEFFICIENTS" not in model_settings:
        logger.warning(
            f"no coefficient file specified in model_settings for {spec_file_name}"
        )
        try:
            assert (spec.astype(float) == spec).all(axis=None)
        except (ValueError, AssertionError):
            raise RuntimeError(
                f"No coefficient file specified for {spec_file_name} "
                f"but not all spec column values are numeric"
            )

        return spec

    coefficients = read_model_coefficients(model_settings)

    spec = eval_coefficients(spec, coefficients, estimator)

    return spec


def read_model_coefficient_template(model_settings):
    """
    Read the coefficient template specified by COEFFICIENT_TEMPLATE model setting
    """

    assert (
        "COEFFICIENT_TEMPLATE" in model_settings
    ), "'COEFFICIENT_TEMPLATE' not in model_settings in %s" % model_settings.get(
        "source_file_paths"
    )

    coefficients_file_name = model_settings["COEFFICIENT_TEMPLATE"]

    file_path = config.config_file_path(coefficients_file_name)
    try:
        template = pd.read_csv(file_path, comment="#", index_col="coefficient_name")
    except ValueError:
        logger.exception("Coefficient Template File Invalid: %s" % str(file_path))
        raise

    # by convention, an empty cell in the template indicates that
    # the coefficient name should be propogated to across all segments
    # this makes for a more legible template than repeating the identical coefficient name in each column

    # replace missing cell values with coefficient_name from index
    template = template.where(
        ~template.isnull(),
        np.broadcast_to(template.index.values[:, None], template.shape),
    )

    if template.index.duplicated().any():
        dupes = template[template.index.duplicated(keep=False)].sort_index()
        logger.warning(
            f"duplicate coefficient names in {coefficients_file_name}:\n{dupes}"
        )
        assert not template.index.duplicated().any()

    return template


def dump_mapped_coefficients(model_settings):
    """
    dump template_df with coefficient values
    """

    coefficients_df = read_model_coefficients(model_settings)
    template_df = read_model_coefficient_template(model_settings)

    for c in template_df.columns:
        template_df[c] = template_df[c].map(coefficients_df.value)

    coefficients_template_file_name = model_settings["COEFFICIENT_TEMPLATE"]
    file_path = config.output_file_path(coefficients_template_file_name)
    template_df.to_csv(file_path, index=True)
    logger.info(f"wrote mapped coefficient template to {file_path}")

    coefficients_file_name = model_settings["COEFFICIENTS"]
    file_path = config.output_file_path(coefficients_file_name)
    coefficients_df.to_csv(file_path, index=True)
    logger.info(f"wrote raw coefficients to {file_path}")


def get_segment_coefficients(model_settings, segment_name):
    """
    Return a dict mapping generic coefficient names to segment-specific coefficient values

    some specs mode_choice logsums have the same espression values with different coefficients for various segments
    (e.g. eatout, .. ,atwork) and a template file that maps a flat list of coefficients into segment columns.

    This allows us to provide a coefficient file with just the coefficients for a specific segment,
    that works with generic coefficient names in the spec. For instance coef_ivt can take on the values
    of segment-specific coefficients coef_ivt_school_univ, coef_ivt_work, coef_ivt_atwork,...

    ::

        coefficients_df
                                      value constrain
        coefficient_name
        coef_ivt_eatout_escort_...  -0.0175         F
        coef_ivt_school_univ        -0.0224         F
        coef_ivt_work               -0.0134         F
        coef_ivt_atwork             -0.0188         F

        template_df

        coefficient_name     eatout                       school                 school                 work
        coef_ivt             coef_ivt_eatout_escort_...   coef_ivt_school_univ   coef_ivt_school_univ   coef_ivt_work

        For school segment this will return the generic coefficient name with the segment-specific coefficient value
        e.g. {'coef_ivt': -0.0224, ...}
        ...

    """

    if "COEFFICIENTS" in model_settings and "COEFFICIENT_TEMPLATE" in model_settings:
        legacy = False
    elif "COEFFICIENTS" in model_settings:
        legacy = "COEFFICIENTS"
        warnings.warn(
            "Support for COEFFICIENTS without COEFFICIENT_TEMPLATE in model settings file will be removed."
            "Use COEFFICIENT and COEFFICIENT_TEMPLATE to support estimation.",
            FutureWarning,
        )
    elif "LEGACY_COEFFICIENTS" in model_settings:
        legacy = "LEGACY_COEFFICIENTS"
        warnings.warn(
            "Support for 'LEGACY_COEFFICIENTS' setting in model settings file will be removed."
            "Use COEFFICIENT and COEFFICIENT_TEMPLATE to support estimation.",
            FutureWarning,
        )
    else:
        raise RuntimeError(f"No COEFFICIENTS setting in model_settings")

    if legacy:
        constants = config.get_model_constants(model_settings)
        legacy_coeffs_file_path = config.config_file_path(model_settings[legacy])
        omnibus_coefficients = pd.read_csv(
            legacy_coeffs_file_path, comment="#", index_col="coefficient_name"
        )
        coefficients_dict = assign.evaluate_constants(
            omnibus_coefficients[segment_name], constants=constants
        )
    else:
        coefficients_df = read_model_coefficients(model_settings)
        template_df = read_model_coefficient_template(model_settings)
        coefficients_col = (
            template_df[segment_name].map(coefficients_df.value).astype(float)
        )

        if coefficients_col.isnull().any():
            # show them the offending lines from interaction_coefficients_file
            logger.warning(
                f"bad coefficients in COEFFICIENTS {model_settings['COEFFICIENTS']}\n"
                f"{coefficients_col[coefficients_col.isnull()]}"
            )
            assert not coefficients_col.isnull().any()

        coefficients_dict = coefficients_col.to_dict()

    return coefficients_dict


def eval_nest_coefficients(nest_spec, coefficients, trace_label):
    def replace_coefficients(nest):
        if isinstance(nest, dict):

            assert "coefficient" in nest
            coefficient_name = nest["coefficient"]
            if isinstance(coefficient_name, str):
                assert (
                    coefficient_name in coefficients
                ), "%s not in nest coefficients" % (coefficient_name,)
                nest["coefficient"] = coefficients[coefficient_name]

            assert "alternatives" in nest
            for alternative in nest["alternatives"]:
                if isinstance(alternative, dict):
                    replace_coefficients(alternative)

    if isinstance(coefficients, pd.DataFrame):
        assert "value" in coefficients.columns
        coefficients = coefficients["value"].to_dict()

    replace_coefficients(nest_spec)

    logit.validate_nest_spec(nest_spec, trace_label)

    return nest_spec


def eval_coefficients(spec, coefficients, estimator):

    spec = spec.copy()  # don't clobber input spec

    if isinstance(coefficients, pd.DataFrame):
        assert "value" in coefficients.columns
        coefficients = coefficients["value"].to_dict()

    assert isinstance(
        coefficients, dict
    ), "eval_coefficients doesn't grok type of coefficients: %s" % (type(coefficients))

    for c in spec.columns:
        if c == SPEC_LABEL_NAME:
            continue
        spec[c] = (
            spec[c].apply(lambda x: eval(str(x), {}, coefficients)).astype(np.float32)
        )

    # drop any rows with all zeros since they won't have any effect (0 marginal utility)
    # (do not drop rows in estimation mode as it may confuse the estimation package (e.g. larch)
    zero_rows = (spec == 0).all(axis=1)
    if zero_rows.any():
        if estimator:
            logger.debug("keeping %s all-zero rows in SPEC" % (zero_rows.sum(),))
        else:
            logger.debug("dropping %s all-zero rows from SPEC" % (zero_rows.sum(),))
            spec = spec.loc[~zero_rows]

    return spec


def eval_utilities(
    spec,
    choosers,
    locals_d=None,
    trace_label=None,
    have_trace_targets=False,
    trace_all_rows=False,
    estimator=None,
    trace_column_names=None,
    log_alt_losers=False,
):
    """

    Parameters
    ----------
    spec : pandas.DataFrame
        A table of variable specifications and coefficient values.
        Variable expressions should be in the table index and the table
        should have a column for each alternative.
    choosers : pandas.DataFrame
    locals_d : Dict or None
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    trace_label: str
    have_trace_targets: boolean - choosers has targets to trace
    trace_all_rows: boolean - trace all chooser rows, bypassing tracing.trace_targets
    estimator :
        called to report intermediate table results (used for estimation)
    trace_column_names: str or list of str
        chooser columns to include when tracing expression_values

    Returns
    -------

    """

    # fixme - restore tracing and _check_for_variability

    trace_label = tracing.extend_trace_label(trace_label, "eval_utils")

    # avoid altering caller's passed-in locals_d parameter (they may be looping)
    locals_dict = assign.local_utilities()

    if locals_d is not None:
        locals_dict.update(locals_d)
    globals_dict = {}

    locals_dict["df"] = choosers

    # - eval spec expressions
    if isinstance(spec.index, pd.MultiIndex):
        # spec MultiIndex with expression and label
        exprs = spec.index.get_level_values(SPEC_EXPRESSION_NAME)
    else:
        exprs = spec.index

    expression_values = np.empty((spec.shape[0], choosers.shape[0]))
    chunk.log_df(trace_label, "expression_values", expression_values)

    i = 0
    for expr, coefficients in zip(exprs, spec.values):

        try:
            with warnings.catch_warnings(record=True) as w:
                # Cause all warnings to always be triggered.
                warnings.simplefilter("always")
                if expr.startswith("@"):
                    expression_value = eval(expr[1:], globals_dict, locals_dict)

                else:
                    expression_value = choosers.eval(expr)

                if len(w) > 0:
                    for wrn in w:
                        logger.warning(
                            f"{trace_label} - {type(wrn).__name__} ({wrn.message}) evaluating: {str(expr)}"
                        )

        except Exception as err:
            logger.exception(
                f"{trace_label} - {type(err).__name__} ({str(err)}) evaluating: {str(expr)}"
            )
            raise err

        if log_alt_losers:
            # utils for each alt for this expression
            # FIXME if we always did tis, we cold uem these and skip np.dot below
            utils = np.outer(expression_value, coefficients)
            losers = np.amax(utils, axis=1) < ALT_LOSER_UTIL

            if losers.any():
                logger.warning(
                    f"{trace_label} - {sum(losers)} choosers of {len(losers)} "
                    f"with prohibitive utilities for all alternatives for expression: {expr}"
                )

        expression_values[i] = expression_value
        i += 1

    chunk.log_df(trace_label, "expression_values", expression_values)

    if estimator:
        df = pd.DataFrame(
            data=expression_values.transpose(),
            index=choosers.index,
            columns=spec.index.get_level_values(SPEC_LABEL_NAME),
        )
        df.index.name = choosers.index.name
        estimator.write_expression_values(df)

    # - compute_utilities
    utilities = np.dot(expression_values.transpose(), spec.astype(np.float64).values)
    utilities = pd.DataFrame(data=utilities, index=choosers.index, columns=spec.columns)

    chunk.log_df(trace_label, "utilities", utilities)

    # sometimes tvpb will drop rows on the fly and we wind up with an empty
    # table of choosers. this will just bypass tracing in that case.
    if (trace_all_rows or have_trace_targets) and (len(choosers) > 0):

        if trace_all_rows:
            trace_targets = pd.Series(True, index=choosers.index)
        else:
            trace_targets = tracing.trace_targets(choosers)
            assert trace_targets.any()  # since they claimed to have targets...

        # get int offsets of the trace_targets (offsets of bool=True values)
        offsets = np.nonzero(list(trace_targets))[0]

        # get array of expression_values
        # expression_values.shape = (len(spec), len(choosers))
        # data.shape = (len(spec), len(offsets))
        data = expression_values[:, offsets]

        # index is utility expressions (and optional label if MultiIndex)
        expression_values_df = pd.DataFrame(data=data, index=spec.index)

        if trace_column_names is not None:
            if isinstance(trace_column_names, str):
                trace_column_names = [trace_column_names]
            expression_values_df.columns = pd.MultiIndex.from_frame(
                choosers.loc[trace_targets, trace_column_names]
            )

        tracing.trace_df(
            expression_values_df,
            tracing.extend_trace_label(trace_label, "expression_values"),
            slicer=None,
            transpose=False,
        )

        if len(spec.columns) > 1:

            for c in spec.columns:
                name = f"expression_value_{c}"

                tracing.trace_df(
                    expression_values_df.multiply(spec[c].values, axis=0),
                    tracing.extend_trace_label(trace_label, name),
                    slicer=None,
                    transpose=False,
                )

    del expression_values
    chunk.log_df(trace_label, "expression_values", None)

    # no longer our problem - but our caller should re-log this...
    chunk.log_df(trace_label, "utilities", None)

    return utilities


def eval_variables(exprs, df, locals_d=None):
    """
    Evaluate a set of variable expressions from a spec in the context
    of a given data table.

    There are two kinds of supported expressions: "simple" expressions are
    evaluated in the context of the DataFrame using DataFrame.eval.
    This is the default type of expression.

    Python expressions are evaluated in the context of this function using
    Python's eval function. Because we use Python's eval this type of
    expression supports more complex operations than a simple expression.
    Python expressions are denoted by beginning with the @ character.
    Users should take care that these expressions must result in
    a Pandas Series.

    # FIXME - for performance, it is essential that spec and expression_values
    # FIXME - not contain booleans when dotted with spec values
    # FIXME - or the arrays will be converted to dtype=object within dot()

    Parameters
    ----------
    exprs : sequence of str
    df : pandas.DataFrame
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @

    Returns
    -------
    variables : pandas.DataFrame
        Will have the index of `df` and columns of eval results of `exprs`.
    """

    # avoid altering caller's passed-in locals_d parameter (they may be looping)
    locals_dict = assign.local_utilities()
    if locals_d is not None:
        locals_dict.update(locals_d)
    globals_dict = {}

    locals_dict["df"] = df

    def to_array(x):

        if x is None or np.isscalar(x):
            a = np.asanyarray([x] * len(df.index))
        elif isinstance(x, pd.Series):
            # fixme
            # assert x.index.equals(df.index)
            # save a little RAM
            a = x.values
        else:
            a = x

        # FIXME - for performance, it is essential that spec and expression_values
        # FIXME - not contain booleans when dotted with spec values
        # FIXME - or the arrays will be converted to dtype=object within dot()
        if not np.issubdtype(a.dtype, np.number):
            a = a.astype(np.int8)

        return a

    values = OrderedDict()
    for expr in exprs:
        try:
            if expr.startswith("@"):
                expr_values = to_array(eval(expr[1:], globals_dict, locals_dict))
            else:
                expr_values = to_array(df.eval(expr))
            # read model spec should ensure uniqueness, otherwise we should uniquify
            assert expr not in values
            values[expr] = expr_values

        except Exception as err:
            logger.exception(
                f"Variable evaluation failed {type(err).__name__} ({str(err)}) evaluating: {str(expr)}"
            )
            raise err

    values = util.df_from_dict(values, index=df.index)

    return values


# no longer used because eval_utilities aggregates expression_values as they are computed to save space
# def compute_utilities(expression_values, spec):
#
#     # matrix product of spec expression_values with utility coefficients of alternatives
#     # sums the partial utilities (represented by each spec row) of the alternatives
#     # resulting in a dataframe with one row per chooser and one column per alternative
#     # pandas.dot depends on column names of expression_values matching spec index values
#
#     # FIXME - for performance, it is essential that spec and expression_values
#     # FIXME - not contain booleans when dotted with spec values
#     # FIXME - or the arrays will be converted to dtype=object within dot()
#
#     spec = spec.astype(np.float64)
#
#     # pandas.dot depends on column names of expression_values matching spec index values
#     # expressions should have been uniquified when spec was read
#     # we could do it here if need be, and then set spec.index and expression_values.columns equal
#     assert spec.index.is_unique
#     assert (spec.index.values == expression_values.columns.values).all()
#
#     utilities = expression_values.dot(spec)
#
#     return utilities


def set_skim_wrapper_targets(df, skims):
    """
    Add the dataframe to the SkimWrapper object so that it can be dereferenced
    using the parameters of the skims object.

    Parameters
    ----------
    df : pandas.DataFrame
        Table to which to add skim data as new columns.
        `df` is modified in-place.
    skims : SkimWrapper or Skim3dWrapper object, or a list or dict of skims
        The skims object is used to contain multiple matrices of
        origin-destination impedances.  Make sure to also add it to the
        locals_d below in order to access it in expressions.  The *only* job
        of this method in regards to skims is to call set_df with the
        dataframe that comes back from interacting choosers with
        alternatives.  See the skims module for more documentation on how
        the skims object is intended to be used.
    """

    skims = (
        skims
        if isinstance(skims, list)
        else skims.values()
        if isinstance(skims, dict)
        else [skims]
    )

    # assume any object in skims can be treated as a skim
    for skim in skims:
        try:
            skim.set_df(df)
        except AttributeError:
            pass


def _check_for_variability(expression_values, trace_label):
    """
    This is an internal method which checks for variability in each
    expression - under the assumption that you probably wouldn't be using a
    variable (in live simulations) if it had no variability.  This is a
    warning to the user that they might have constructed the variable
    incorrectly.  It samples 1000 rows in order to not hurt performance -
    it's likely that if 1000 rows have no variability, the whole dataframe
    will have no variability.
    """

    if trace_label is None:
        trace_label = "_check_for_variability"

    sample = random_rows(expression_values, min(1000, len(expression_values)))

    no_variability = has_missing_vals = 0
    for i in range(len(sample.columns)):
        v = sample.iloc[:, i]
        if v.min() == v.max():
            col_name = sample.columns[i]
            logger.info(
                "%s: no variability (%s) in: %s" % (trace_label, v.iloc[0], col_name)
            )
            no_variability += 1
        # FIXME - how could this happen? Not sure it is really a problem?
        if np.count_nonzero(v.isnull().values) > 0:
            col_name = sample.columns[i]
            logger.info("%s: missing values in: %s" % (trace_label, col_name))
            has_missing_vals += 1

    if no_variability > 0:
        logger.warning(
            "%s: %s columns have no variability" % (trace_label, no_variability)
        )

    if has_missing_vals > 0:
        logger.warning(
            "%s: %s columns have missing values" % (trace_label, has_missing_vals)
        )


def compute_nested_exp_utilities(raw_utilities, nest_spec):
    """
    compute exponentiated nest utilities based on nesting coefficients

    For nest nodes this is the exponentiated logsum of alternatives adjusted by nesting coefficient

    leaf <- exp( raw_utility )
    nest <- exp( ln(sum of exponentiated raw_utility of leaves) * nest_coefficient)

    Parameters
    ----------
    raw_utilities : pandas.DataFrame
        dataframe with the raw alternative utilities of all leaves
        (what in non-nested logit would be the utilities of all the alternatives)
    nest_spec : dict
        Nest tree dict from the model spec yaml file

    Returns
    -------
    nested_utilities : pandas.DataFrame
        Will have the index of `raw_utilities` and columns for exponentiated leaf and node utilities
    """
    nested_utilities = pd.DataFrame(index=raw_utilities.index)

    for nest in logit.each_nest(nest_spec, post_order=True):

        name = nest.name

        if nest.is_leaf:
            # leaf_utility = raw_utility / nest.product_of_coefficients
            nested_utilities[name] = (
                raw_utilities[name].astype(float) / nest.product_of_coefficients
            )

        else:
            # nest node
            # the alternative nested_utilities will already have been computed due to post_order
            # this will RuntimeWarning: divide by zero encountered in log
            # if all nest alternative utilities are zero
            # but the resulting inf will become 0 when exp is applied below
            with np.errstate(divide="ignore"):
                nested_utilities[name] = nest.coefficient * np.log(
                    nested_utilities[nest.alternatives].sum(axis=1)
                )

        # exponentiate the utility
        nested_utilities[name] = np.exp(nested_utilities[name])

    return nested_utilities


def compute_nested_probabilities(nested_exp_utilities, nest_spec, trace_label):
    """
    compute nested probabilities for nest leafs and nodes
    probability for nest alternatives is simply the alternatives's local (to nest) probability
    computed in the same way as the probability of non-nested alternatives in multinomial logit
    i.e. the fractional share of the sum of the exponentiated utility of itself and its siblings
    except in nested logit, its sib group is restricted to the nest

    Parameters
    ----------
    nested_exp_utilities : pandas.DataFrame
        dataframe with the exponentiated nested utilities of all leaves and nodes
    nest_spec : dict
        Nest tree dict from the model spec yaml file
    Returns
    -------
    nested_probabilities : pandas.DataFrame
        Will have the index of `nested_exp_utilities` and columns for leaf and node probabilities
    """

    nested_probabilities = pd.DataFrame(index=nested_exp_utilities.index)

    for nest in logit.each_nest(nest_spec, type="node", post_order=False):

        probs = logit.utils_to_probs(
            nested_exp_utilities[nest.alternatives],
            trace_label=trace_label,
            exponentiated=True,
            allow_zero_probs=True,
        )

        nested_probabilities = pd.concat([nested_probabilities, probs], axis=1)

    return nested_probabilities


def compute_base_probabilities(nested_probabilities, nests, spec):
    """
    compute base probabilities for nest leaves
    Base probabilities will be the nest-adjusted probabilities of all leaves
    This flattens or normalizes all the nested probabilities so that they have the proper global
    relative values (the leaf probabilities sum to 1 for each row.)

    Parameters
    ----------
    nested_probabilities : pandas.DataFrame
        dataframe with the nested probabilities for nest leafs and nodes
    nests : dict
        Nest tree dict from the model spec yaml file
    spec : pandas.Dataframe
        simple simulate spec so we can return columns in appropriate order
    Returns
    -------
    base_probabilities : pandas.DataFrame
        Will have the index of `nested_probabilities` and columns for leaf base probabilities
    """

    base_probabilities = pd.DataFrame(index=nested_probabilities.index)

    for nest in logit.each_nest(nests, type="leaf", post_order=False):

        # skip root: it has a prob of 1 but we didn't compute a nested probability column for it
        ancestors = nest.ancestors[1:]

        base_probabilities[nest.name] = nested_probabilities[ancestors].prod(axis=1)

    # reorder alternative columns to match spec
    # since these are alternatives chosen by column index, order of columns matters
    assert set(base_probabilities.columns) == set(spec.columns)
    base_probabilities = base_probabilities[spec.columns]

    return base_probabilities


def eval_mnl(
    choosers,
    spec,
    locals_d,
    custom_chooser,
    estimator,
    log_alt_losers=False,
    want_logsums=False,
    trace_label=None,
    trace_choice_name=None,
    trace_column_names=None,
):
    """
    Run a simulation for when the model spec does not involve alternative
    specific data, e.g. there are no interactions with alternative
    properties and no need to sample from alternatives.

    Each row in spec computes a partial utility for each alternative,
    by providing a spec expression (often a boolean 0-1 trigger)
    and a column of utility coefficients for each alternative.

    We compute the utility of each alternative by matrix-multiplication of eval results
    with the utility coefficients in the spec alternative columns
    yielding one row per chooser and one column per alternative

    Parameters
    ----------
    choosers : pandas.DataFrame
    spec : pandas.DataFrame
        A table of variable specifications and coefficient values.
        Variable expressions should be in the table index and the table
        should have a column for each alternative.
    locals_d : Dict or None
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    custom_chooser : function(probs, choosers, spec, trace_label) returns choices, rands
        custom alternative to logit.make_choices
    estimator : Estimator object
        called to report intermediate table results (used for estimation)
    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.
    trace_choice_name: str
        This is the column label to be used in trace file csv dump of choices
    trace_column_names: str or list of str
        chooser columns to include when tracing expression_values

    Returns
    -------
    choices : pandas.Series
        Index will be that of `choosers`, values will match the columns
        of `spec`.
    """

    # FIXME - not implemented because not currently needed
    assert not want_logsums

    trace_label = tracing.extend_trace_label(trace_label, "eval_mnl")
    have_trace_targets = tracing.has_trace_targets(choosers)

    if have_trace_targets:
        tracing.trace_df(choosers, "%s.choosers" % trace_label)

    utilities = eval_utilities(
        spec,
        choosers,
        locals_d,
        log_alt_losers=log_alt_losers,
        trace_label=trace_label,
        have_trace_targets=have_trace_targets,
        estimator=estimator,
        trace_column_names=trace_column_names,
    )
    chunk.log_df(trace_label, "utilities", utilities)

    if have_trace_targets:
        tracing.trace_df(
            utilities,
            "%s.utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )

    probs = logit.utils_to_probs(
        utilities, trace_label=trace_label, trace_choosers=choosers
    )
    chunk.log_df(trace_label, "probs", probs)

    del utilities
    chunk.log_df(trace_label, "utilities", None)

    if have_trace_targets:
        # report these now in case make_choices throws error on bad_choices
        tracing.trace_df(
            probs,
            "%s.probs" % trace_label,
            column_labels=["alternative", "probability"],
        )

    if custom_chooser:
        choices, rands = custom_chooser(
            probs=probs, choosers=choosers, spec=spec, trace_label=trace_label
        )
    else:
        choices, rands = logit.make_choices(probs, trace_label=trace_label)

    del probs
    chunk.log_df(trace_label, "probs", None)

    if have_trace_targets:
        tracing.trace_df(
            choices, "%s.choices" % trace_label, columns=[None, trace_choice_name]
        )
        tracing.trace_df(rands, "%s.rands" % trace_label, columns=[None, "rand"])

    return choices


def eval_nl(
    choosers,
    spec,
    nest_spec,
    locals_d,
    custom_chooser,
    estimator,
    log_alt_losers=False,
    want_logsums=False,
    trace_label=None,
    trace_choice_name=None,
    trace_column_names=None,
):
    """
    Run a nested-logit simulation for when the model spec does not involve alternative
    specific data, e.g. there are no interactions with alternative
    properties and no need to sample from alternatives.

    Parameters
    ----------
    choosers : pandas.DataFrame
    spec : pandas.DataFrame
        A table of variable specifications and coefficient values.
        Variable expressions should be in the table index and the table
        should have a column for each alternative.
    nest_spec:
        dictionary specifying nesting structure and nesting coefficients
        (from the model spec yaml file)
    locals_d : Dict or None
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    custom_chooser : function(probs, choosers, spec, trace_label) returns choices, rands
        custom alternative to logit.make_choices
    estimator : Estimator object
        called to report intermediate table results (used for estimation)
    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.
    trace_choice_name: str
        This is the column label to be used in trace file csv dump of choices
    trace_column_names: str or list of str
        chooser columns to include when tracing expression_values

    Returns
    -------
    choices : pandas.Series
        Index will be that of `choosers`, values will match the columns
        of `spec`.
    """

    trace_label = tracing.extend_trace_label(trace_label, "eval_nl")
    assert trace_label
    have_trace_targets = tracing.has_trace_targets(choosers)

    logit.validate_nest_spec(nest_spec, trace_label)

    if have_trace_targets:
        tracing.trace_df(choosers, "%s.choosers" % trace_label)

    raw_utilities = eval_utilities(
        spec,
        choosers,
        locals_d,
        log_alt_losers=log_alt_losers,
        trace_label=trace_label,
        have_trace_targets=have_trace_targets,
        estimator=estimator,
        trace_column_names=trace_column_names,
    )
    chunk.log_df(trace_label, "raw_utilities", raw_utilities)

    if have_trace_targets:
        tracing.trace_df(
            raw_utilities,
            "%s.raw_utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )

    # exponentiated utilities of leaves and nests
    nested_exp_utilities = compute_nested_exp_utilities(raw_utilities, nest_spec)
    chunk.log_df(trace_label, "nested_exp_utilities", nested_exp_utilities)

    del raw_utilities
    chunk.log_df(trace_label, "raw_utilities", None)

    if have_trace_targets:
        tracing.trace_df(
            nested_exp_utilities,
            "%s.nested_exp_utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )

    # probabilities of alternatives relative to siblings sharing the same nest
    nested_probabilities = compute_nested_probabilities(
        nested_exp_utilities, nest_spec, trace_label=trace_label
    )
    chunk.log_df(trace_label, "nested_probabilities", nested_probabilities)

    if want_logsums:
        # logsum of nest root
        logsums = pd.Series(np.log(nested_exp_utilities.root), index=choosers.index)
        chunk.log_df(trace_label, "logsums", logsums)

    del nested_exp_utilities
    chunk.log_df(trace_label, "nested_exp_utilities", None)

    if have_trace_targets:
        tracing.trace_df(
            nested_probabilities,
            "%s.nested_probabilities" % trace_label,
            column_labels=["alternative", "probability"],
        )

    # global (flattened) leaf probabilities based on relative nest coefficients (in spec order)
    base_probabilities = compute_base_probabilities(
        nested_probabilities, nest_spec, spec
    )
    chunk.log_df(trace_label, "base_probabilities", base_probabilities)

    del nested_probabilities
    chunk.log_df(trace_label, "nested_probabilities", None)

    if have_trace_targets:
        tracing.trace_df(
            base_probabilities,
            "%s.base_probabilities" % trace_label,
            column_labels=["alternative", "probability"],
        )

    # note base_probabilities could all be zero since we allowed all probs for nests to be zero
    # check here to print a clear message but make_choices will raise error if probs don't sum to 1
    BAD_PROB_THRESHOLD = 0.001
    no_choices = (base_probabilities.sum(axis=1) - 1).abs() > BAD_PROB_THRESHOLD

    if no_choices.any():

        logit.report_bad_choices(
            no_choices,
            base_probabilities,
            trace_label=tracing.extend_trace_label(trace_label, "bad_probs"),
            trace_choosers=choosers,
            msg="base_probabilities do not sum to one",
        )

    if custom_chooser:
        choices, rands = custom_chooser(
            probs=base_probabilities,
            choosers=choosers,
            spec=spec,
            trace_label=trace_label,
        )
    else:
        choices, rands = logit.make_choices(base_probabilities, trace_label=trace_label)

    del base_probabilities
    chunk.log_df(trace_label, "base_probabilities", None)

    if have_trace_targets:
        tracing.trace_df(
            choices, "%s.choices" % trace_label, columns=[None, trace_choice_name]
        )
        tracing.trace_df(rands, "%s.rands" % trace_label, columns=[None, "rand"])
        if want_logsums:
            tracing.trace_df(
                logsums, "%s.logsums" % trace_label, columns=[None, "logsum"]
            )

    if want_logsums:
        choices = choices.to_frame("choice")
        choices["logsum"] = logsums

    return choices


def _simple_simulate(
    choosers,
    spec,
    nest_spec,
    skims=None,
    locals_d=None,
    custom_chooser=None,
    log_alt_losers=False,
    want_logsums=False,
    estimator=None,
    trace_label=None,
    trace_choice_name=None,
    trace_column_names=None,
):
    """
    Run an MNL or NL simulation for when the model spec does not involve alternative
    specific data, e.g. there are no interactions with alternative
    properties and no need to sample from alternatives.

    Parameters
    ----------
    choosers : pandas.DataFrame
    spec : pandas.DataFrame
        A table of variable specifications and coefficient values.
        Variable expressions should be in the table index and the table
        should have a column for each alternative.
    nest_spec:
        for nested logit (nl): dictionary specifying nesting structure and nesting coefficients
        for multinomial logit (mnl): None
    skims : Skims object
        The skims object is used to contain multiple matrices of
        origin-destination impedances.  Make sure to also add it to the
        locals_d below in order to access it in expressions.  The *only* job
        of this method in regards to skims is to call set_df with the
        dataframe that comes back from interacting choosers with
        alternatives.  See the skims module for more documentation on how
        the skims object is intended to be used.
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    custom_chooser : Estimator object
    estimator : function(df, label, table_name)
        called to report intermediate table results (used for estimation)

    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.
    trace_choice_name: str
        This is the column label to be used in trace file csv dump of choices
    trace_column_names: str or list of str
        chooser columns to include when tracing expression_values

    Returns
    -------
    choices : pandas.Series
        Index will be that of `choosers`, values will match the columns
        of `spec`.
    """

    if skims is not None:
        set_skim_wrapper_targets(choosers, skims)

    if nest_spec is None:
        choices = eval_mnl(
            choosers,
            spec,
            locals_d,
            custom_chooser,
            log_alt_losers=log_alt_losers,
            want_logsums=want_logsums,
            estimator=estimator,
            trace_label=trace_label,
            trace_choice_name=trace_choice_name,
            trace_column_names=trace_column_names,
        )
    else:
        choices = eval_nl(
            choosers,
            spec,
            nest_spec,
            locals_d,
            custom_chooser,
            log_alt_losers=log_alt_losers,
            want_logsums=want_logsums,
            estimator=estimator,
            trace_label=trace_label,
            trace_choice_name=trace_choice_name,
            trace_column_names=trace_column_names,
        )

    return choices


def tvpb_skims(skims):
    def list_of_skims(skims):
        return (
            skims
            if isinstance(skims, list)
            else skims.values()
            if isinstance(skims, dict)
            else [skims]
            if skims is not None
            else []
        )

    return [
        skim
        for skim in list_of_skims(skims)
        if isinstance(skim, pathbuilder.TransitVirtualPathLogsumWrapper)
    ]


def simple_simulate(
    choosers,
    spec,
    nest_spec,
    skims=None,
    locals_d=None,
    chunk_size=0,
    custom_chooser=None,
    log_alt_losers=False,
    want_logsums=False,
    estimator=None,
    trace_label=None,
    trace_choice_name=None,
    trace_column_names=None,
):
    """
    Run an MNL or NL simulation for when the model spec does not involve alternative
    specific data, e.g. there are no interactions with alternative
    properties and no need to sample from alternatives.
    """

    trace_label = tracing.extend_trace_label(trace_label, "simple_simulate")

    assert len(choosers) > 0

    result_list = []
    # segment by person type and pick the right spec for each person type
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        choosers, chunk_size, trace_label
    ):

        choices = _simple_simulate(
            chooser_chunk,
            spec,
            nest_spec,
            skims=skims,
            locals_d=locals_d,
            custom_chooser=custom_chooser,
            log_alt_losers=log_alt_losers,
            want_logsums=want_logsums,
            estimator=estimator,
            trace_label=chunk_trace_label,
            trace_choice_name=trace_choice_name,
            trace_column_names=trace_column_names,
        )

        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert len(choices.index == len(choosers.index))

    return choices


def simple_simulate_by_chunk_id(
    choosers,
    spec,
    nest_spec,
    skims=None,
    locals_d=None,
    chunk_size=0,
    custom_chooser=None,
    log_alt_losers=False,
    want_logsums=False,
    estimator=None,
    trace_label=None,
    trace_choice_name=None,
):
    """
    chunk_by_chunk_id wrapper for simple_simulate
    """

    result_list = []
    for (
        i,
        chooser_chunk,
        chunk_trace_label,
    ) in chunk.adaptive_chunked_choosers_by_chunk_id(choosers, chunk_size, trace_label):

        choices = _simple_simulate(
            chooser_chunk,
            spec,
            nest_spec,
            skims=skims,
            locals_d=locals_d,
            custom_chooser=custom_chooser,
            log_alt_losers=log_alt_losers,
            want_logsums=want_logsums,
            estimator=estimator,
            trace_label=chunk_trace_label,
            trace_choice_name=trace_choice_name,
        )

        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

    if len(result_list) > 1:
        choices = pd.concat(result_list)

    return choices


def eval_mnl_logsums(choosers, spec, locals_d, trace_label=None):
    """
    like eval_nl except return logsums instead of making choices

    Returns
    -------
    logsums : pandas.Series
        Index will be that of `choosers`, values will be logsum across spec column values
    """

    # FIXME - untested and not currently used by any models...

    trace_label = tracing.extend_trace_label(trace_label, "eval_mnl_logsums")
    have_trace_targets = tracing.has_trace_targets(choosers)

    logger.debug("running eval_mnl_logsums")

    # trace choosers
    if have_trace_targets:
        tracing.trace_df(choosers, "%s.choosers" % trace_label)

    utilities = eval_utilities(
        spec, choosers, locals_d, trace_label, have_trace_targets
    )
    chunk.log_df(trace_label, "utilities", utilities)

    if have_trace_targets:
        tracing.trace_df(
            utilities,
            "%s.raw_utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )

    # - logsums
    # logsum is log of exponentiated utilities summed across columns of each chooser row
    logsums = np.log(np.exp(utilities.values).sum(axis=1))
    logsums = pd.Series(logsums, index=choosers.index)
    chunk.log_df(trace_label, "logsums", logsums)

    # trace utilities
    if have_trace_targets:
        tracing.trace_df(
            logsums, "%s.logsums" % trace_label, column_labels=["alternative", "logsum"]
        )

    return logsums


def eval_nl_logsums(choosers, spec, nest_spec, locals_d, trace_label=None):
    """
    like eval_nl except return logsums instead of making choices

    Returns
    -------
    logsums : pandas.Series
        Index will be that of `choosers`, values will be nest logsum based on spec column values
    """

    trace_label = tracing.extend_trace_label(trace_label, "eval_nl_logsums")
    have_trace_targets = tracing.has_trace_targets(choosers)

    logit.validate_nest_spec(nest_spec, trace_label)

    # trace choosers
    if have_trace_targets:
        tracing.trace_df(choosers, "%s.choosers" % trace_label)

    raw_utilities = eval_utilities(
        spec,
        choosers,
        locals_d,
        trace_label=trace_label,
        have_trace_targets=have_trace_targets,
    )
    chunk.log_df(trace_label, "raw_utilities", raw_utilities)

    if have_trace_targets:
        tracing.trace_df(
            raw_utilities,
            "%s.raw_utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )

    # - exponentiated utilities of leaves and nests
    nested_exp_utilities = compute_nested_exp_utilities(raw_utilities, nest_spec)
    chunk.log_df(trace_label, "nested_exp_utilities", nested_exp_utilities)

    del raw_utilities  # done with raw_utilities
    chunk.log_df(trace_label, "raw_utilities", None)

    # - logsums
    logsums = np.log(nested_exp_utilities.root)
    logsums = pd.Series(logsums, index=choosers.index)
    chunk.log_df(trace_label, "logsums", logsums)

    if have_trace_targets:
        # add logsum to nested_exp_utilities for tracing
        nested_exp_utilities["logsum"] = logsums
        tracing.trace_df(
            nested_exp_utilities,
            "%s.nested_exp_utilities" % trace_label,
            column_labels=["alternative", "utility"],
        )
        tracing.trace_df(
            logsums, "%s.logsums" % trace_label, column_labels=["alternative", "logsum"]
        )

    del nested_exp_utilities  # done with nested_exp_utilities
    chunk.log_df(trace_label, "nested_exp_utilities", None)

    return logsums


def _simple_simulate_logsums(
    choosers, spec, nest_spec, skims=None, locals_d=None, trace_label=None
):
    """
    like simple_simulate except return logsums instead of making choices

    Returns
    -------
    logsums : pandas.Series
        Index will be that of `choosers`, values will be nest logsum based on spec column values
    """

    if skims is not None:
        set_skim_wrapper_targets(choosers, skims)

    if nest_spec is None:
        logsums = eval_mnl_logsums(choosers, spec, locals_d, trace_label=trace_label)
    else:
        logsums = eval_nl_logsums(
            choosers, spec, nest_spec, locals_d, trace_label=trace_label
        )

    return logsums


def simple_simulate_logsums(
    choosers,
    spec,
    nest_spec,
    skims=None,
    locals_d=None,
    chunk_size=0,
    trace_label=None,
    chunk_tag=None,
):
    """
    like simple_simulate except return logsums instead of making choices

    Returns
    -------
    logsums : pandas.Series
        Index will be that of `choosers`, values will be nest logsum based on spec column values
    """

    assert len(choosers) > 0
    chunk_tag = chunk_tag or trace_label

    result_list = []
    # segment by person type and pick the right spec for each person type
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        choosers, chunk_size, trace_label, chunk_tag
    ):

        logsums = _simple_simulate_logsums(
            chooser_chunk, spec, nest_spec, skims, locals_d, chunk_trace_label
        )

        result_list.append(logsums)

        chunk.log_df(trace_label, f"result_list", result_list)

    if len(result_list) > 1:
        logsums = pd.concat(result_list)

    assert len(logsums.index == len(choosers.index))

    return logsums



COMPONENT_NAMES = [
    # "compute_accessibility",
    "school_location",
    "workplace_location",
    "auto_ownership_simulate",
    "free_parking",
    "cdap_simulate",
    "mandatory_tour_frequency",
    "mandatory_tour_scheduling",
    "joint_tour_frequency",
    "joint_tour_composition",
    "joint_tour_participation",
    "joint_tour_destination",
    "joint_tour_scheduling",
    "non_mandatory_tour_frequency",
    "non_mandatory_tour_destination",
    "non_mandatory_tour_scheduling",
    "tour_mode_choice_simulate",
    "atwork_subtour_frequency",
    "atwork_subtour_destination",
    "atwork_subtour_scheduling",
    "atwork_subtour_mode_choice",
    "stop_frequency",
    "trip_purpose",
    "trip_destination",
    "trip_purpose_and_destination",
    "trip_scheduling",
    "trip_mode_choice",
    # "write_data_dictionary",
    # "track_skim_usage",
    "write_trip_matrices",
    # "write_tables",
]
BENCHMARK_SETTINGS = {
    "households_sample_size": 48_769,  # match hh sample size in example data
}
SKIM_CACHE = False
PRELOAD_INJECTABLES = ("skim_dict",)
REPEAT = 1
NUMBER = 1
TIMEOUT = 36000.0  # ten hours



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import inject, pipeline, tracing

logger = logging.getLogger(__name__)


@inject.table()
def vehicles(households):
    """Creates the vehicles table and load it as an injectable

    This method initializes the `vehicles` table, where the number of rows
    is equal to the sum of `households["auto_ownership"]`.

    Parameters
    ----------
    households :  orca.DataFrameWrapper

    Returns
    -------
    vehicles : pandas.DataFrame
    """

    # initialize vehicles table
    vehicles = households.to_frame().loc[
        households.index.repeat(households["auto_ownership"])
    ]
    vehicles = vehicles.reset_index()[["household_id"]]

    vehicles["vehicle_num"] = vehicles.groupby("household_id").cumcount() + 1
    # tying the vehicle id to the household id in order to ensure reproducability
    vehicles["vehicle_id"] = vehicles.household_id * 10 + vehicles.vehicle_num
    vehicles.set_index("vehicle_id", inplace=True)

    # replace table function with dataframe
    inject.add_table("vehicles", vehicles)

    pipeline.get_rn_generator().add_channel("vehicles", vehicles)
    tracing.register_traceable_table("households", vehicles)

    return vehicles


@inject.table()
def vehicles_merged(vehicles, households_merged):
    """Augments the vehicles table with household attributes

    Parameters
    ----------
    vehicles :  orca.DataFrameWrapper
    households_merged :  orca.DataFrameWrapper

    Returns
    -------
    vehicles_merged : pandas.DataFrame
    """

    vehicles_merged = inject.merge_tables(
        vehicles.name, tables=[vehicles, households_merged]
    )
    return vehicles_merged


inject.broadcast(
    "households_merged", "vehicles", cast_index=True, onto_on="household_id"
)



import numpy as np
import pandas as pd
import pytest

import activitysim.abm.models.trip_departure_choice as tdc
from activitysim.abm.models.util.trip import get_time_windows
from activitysim.core import los

from .setup_utils import setup_dirs


@pytest.fixture(scope="module")
def trips():
    outbound_array = [True, True, False, False, False, True, True, False, False, True]

    trips = pd.DataFrame(
        data={
            "tour_id": [1, 1, 2, 2, 2, 2, 2, 3, 3, 4],
            "trip_duration": [2, 2, 7, 7, 7, 12, 12, 4, 4, 5],
            "inbound_duration": [0, 0, 7, 7, 7, 0, 0, 4, 4, 5],
            "main_leg_duration": [4, 4, 2, 2, 2, 2, 2, 1, 1, 2],
            "outbound_duration": [2, 2, 0, 0, 0, 12, 12, 0, 0, 5],
            "trip_count": [2, 2, 3, 3, 3, 2, 2, 2, 2, 1],
            "trip_num": [1, 2, 1, 2, 3, 1, 2, 1, 2, 1],
            "outbound": outbound_array,
            "chunk_id": [1, 1, 2, 2, 2, 2, 2, 3, 3, 4],
            "is_work": [
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
            ],
            "is_school": [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                False,
            ],
            "is_eatout": [
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
            ],
            "start": [8, 8, 18, 18, 18, 18, 18, 24, 24, 19],
            "end": [14, 14, 39, 39, 39, 39, 39, 29, 29, 26],
            "origin": [3, 5, 15, 12, 24, 8, 17, 8, 9, 6],
            "destination": [5, 9, 12, 24, 20, 17, 18, 9, 11, 14],
        },
        index=range(10),
    )

    trips.index.name = "trip_id"
    return trips


@pytest.fixture(scope="module")
def settings():
    return {
        "skims_file": "skims.omx",
        "skim_time_periods": {"labels": ["EA", "AM", "MD", "PM", "NT"]},
    }


@pytest.fixture(scope="module")
def model_spec():
    index = [
        "@(df['stop_time_duration'] * df['is_work'].astype(int)).astype(int)",
        "@(df['stop_time_duration'] * df['is_school'].astype(int)).astype(int)",
        "@(df['stop_time_duration'] * df['is_eatout'].astype(int)).astype(int)",
    ]

    values = {
        "inbound": [0.933020, 0.370260, 0.994840],
        "outbound": [0.933020, 0.370260, 0.994840],
    }

    return pd.DataFrame(index=index, data=values)


def test_build_patterns(trips):
    time_windows = get_time_windows(48, 3)
    patterns = tdc.build_patterns(trips, time_windows)
    patterns = patterns.sort_values(["tour_id", "outbound", "trip_num"])

    assert patterns.shape[0] == 34
    assert patterns.shape[1] == 6
    assert patterns.index.name == tdc.TOUR_LEG_ID

    output_columns = [
        tdc.TOUR_ID,
        tdc.PATTERN_ID,
        tdc.TRIP_NUM,
        tdc.STOP_TIME_DURATION,
        tdc.TOUR_ID,
        tdc.OUTBOUND,
    ]

    assert set(output_columns).issubset(patterns.columns)


def test_get_tour_legs(trips):
    tour_legs = tdc.get_tour_legs(trips)
    assert tour_legs.index.name == tdc.TOUR_LEG_ID
    assert (
        np.unique(tour_legs[tdc.TOUR_ID].values).shape[0]
        == np.unique(trips[tdc.TOUR_ID].values).shape[0]
    )


def test_generate_alternative(trips):
    alts = tdc.generate_alternatives(trips, tdc.STOP_TIME_DURATION)
    assert alts.shape[0] == 67
    assert alts.shape[1] == 1

    assert alts.index.name == tdc.TRIP_ID
    assert alts.columns[0] == tdc.STOP_TIME_DURATION

    pd.testing.assert_series_equal(
        trips.groupby(trips.index)["trip_duration"].max(),
        alts.groupby(alts.index)[tdc.STOP_TIME_DURATION].max(),
        check_names=False,
    )


def test_apply_stage_two_model(model_spec, trips):
    setup_dirs()
    departures = tdc.apply_stage_two_model(model_spec, trips, 0, "TEST Trip Departure")
    assert len(departures) == len(trips)
    pd.testing.assert_index_equal(departures.index, trips.index)

    departures = pd.concat([trips, departures], axis=1)



# ActivitySim
# See full license in LICENSE.txt.

import os.path

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from .. import inject, logit
from ..simulate import eval_variables


def setup_function():
    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


@pytest.fixture(scope="module")
def data_dir():
    return os.path.join(os.path.dirname(__file__), "data")


def add_canonical_dirs():

    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)


# this is lifted straight from urbansim's test_mnl.py
@pytest.fixture(
    scope="module",
    params=[
        (
            "fish.csv",
            "fish_choosers.csv",
            pd.DataFrame(
                [[-0.02047652], [0.95309824]], index=["price", "catch"], columns=["Alt"]
            ),
            pd.DataFrame(
                [
                    [0.2849598, 0.2742482, 0.1605457, 0.2802463],
                    [0.1498991, 0.4542377, 0.2600969, 0.1357664],
                ],
                columns=["beach", "boat", "charter", "pier"],
            ),
        )
    ],
)
def test_data(request):
    data, choosers, spec, probabilities = request.param
    return {
        "data": data,
        "choosers": choosers,
        "spec": spec,
        "probabilities": probabilities,
    }


@pytest.fixture
def choosers(test_data, data_dir):
    filen = os.path.join(data_dir, test_data["choosers"])
    return pd.read_csv(filen)


@pytest.fixture
def spec(test_data):
    return test_data["spec"]


@pytest.fixture
def utilities(choosers, spec, test_data):
    vars = eval_variables(spec.index, choosers)
    utils = vars.dot(spec).astype("float")
    return pd.DataFrame(
        utils.values.reshape(test_data["probabilities"].shape),
        columns=test_data["probabilities"].columns,
    )


def test_utils_to_probs(utilities, test_data):
    probs = logit.utils_to_probs(utilities, trace_label=None)
    pdt.assert_frame_equal(probs, test_data["probabilities"])


def test_utils_to_probs_raises():

    add_canonical_dirs()

    idx = pd.Index(name="household_id", data=[1])
    with pytest.raises(RuntimeError) as excinfo:
        logit.utils_to_probs(
            pd.DataFrame([[1, 2, np.inf, 3]], index=idx), trace_label=None
        )
    assert "infinite exponentiated utilities" in str(excinfo.value)

    with pytest.raises(RuntimeError) as excinfo:
        logit.utils_to_probs(
            pd.DataFrame([[-999, -999, -999, -999]], index=idx), trace_label=None
        )
    assert "all probabilities are zero" in str(excinfo.value)


def test_make_choices_only_one():
    probs = pd.DataFrame(
        [[1, 0, 0], [0, 1, 0]], columns=["a", "b", "c"], index=["x", "y"]
    )
    choices, rands = logit.make_choices(probs)

    pdt.assert_series_equal(choices, pd.Series([0, 1], index=["x", "y"]))


def test_make_choices_real_probs(utilities):
    probs = logit.utils_to_probs(utilities, trace_label=None)
    choices, rands = logit.make_choices(probs)

    pdt.assert_series_equal(choices, pd.Series([1, 2], index=[0, 1]))


@pytest.fixture(scope="module")
def interaction_choosers():
    return pd.DataFrame({"attr": ["a", "b", "c", "b"]}, index=["w", "x", "y", "z"])


@pytest.fixture(scope="module")
def interaction_alts():
    return pd.DataFrame({"prop": [10, 20, 30, 40]}, index=[1, 2, 3, 4])


def test_interaction_dataset_no_sample(interaction_choosers, interaction_alts):
    expected = pd.DataFrame(
        {
            "attr": ["a"] * 4 + ["b"] * 4 + ["c"] * 4 + ["b"] * 4,
            "prop": [10, 20, 30, 40] * 4,
        },
        index=[1, 2, 3, 4] * 4,
    )

    interacted = logit.interaction_dataset(interaction_choosers, interaction_alts)

    interacted, expected = interacted.align(expected, axis=1)

    print("interacted\n", interacted)
    print("expected\n", expected)
    pdt.assert_frame_equal(interacted, expected)


def test_interaction_dataset_sampled(interaction_choosers, interaction_alts):
    expected = pd.DataFrame(
        {
            "attr": ["a"] * 2 + ["b"] * 2 + ["c"] * 2 + ["b"] * 2,
            "prop": [30, 40, 10, 30, 40, 10, 20, 10],
        },
        index=[3, 4, 1, 3, 4, 1, 2, 1],
    )

    interacted = logit.interaction_dataset(
        interaction_choosers, interaction_alts, sample_size=2
    )

    interacted, expected = interacted.align(expected, axis=1)
    pdt.assert_frame_equal(interacted, expected)



import numpy as np
import pandas as pd
import pytest

from activitysim.abm.models import trip_scheduling_choice as tsc
from activitysim.abm.tables.skims import skim_dict
from activitysim.core import los

from .setup_utils import setup_dirs


@pytest.fixture(scope="module")
def tours():
    tours = pd.DataFrame(
        data={
            "duration": [2, 44, 32, 12, 11, 16],
            "num_outbound_stops": [2, 4, 0, 0, 1, 3],
            "num_inbound_stops": [1, 0, 0, 2, 1, 2],
            "tour_type": ["othdisc"] * 2 + ["eatout"] * 4,
            "origin": [3, 10, 15, 23, 5, 8],
            "destination": [5, 9, 12, 24, 20, 17],
            tsc.LAST_OB_STOP: [1, 3, 0, 0, 12, 14],
            tsc.FIRST_IB_STOP: [2, 0, 0, 4, 6, 20],
        },
        index=range(6),
    )

    tours.index.name = "tour_id"

    tours[tsc.HAS_OB_STOPS] = tours[tsc.NUM_OB_STOPS] >= 1
    tours[tsc.HAS_IB_STOPS] = tours[tsc.NUM_IB_STOPS] >= 1

    return tours


@pytest.fixture(scope="module")
def settings():
    return {"skims_file": "skims.omx", "skim_time_periods": {"labels": ["MD"]}}


@pytest.fixture(scope="module")
def model_spec():
    index = [
        "@(df['main_leg_duration']>df['duration']).astype(int)",
        "@(df['main_leg_duration'] == 0)&(df['tour_type']=='othdiscr')",
        "@(df['main_leg_duration'] == 1)&(df['tour_type']=='othdiscr')",
        "@(df['main_leg_duration'] == 2)&(df['tour_type']=='othdiscr')",
        "@(df['main_leg_duration'] == 3)&(df['tour_type']=='othdiscr')",
        "@(df['main_leg_duration'] == 4)&(df['tour_type']=='othdiscr')",
        "@df['tour_type']=='othdiscr'",
        "@df['tour_type']=='eatout'",
        "@df['tour_type']=='eatout'",
    ]

    values = [
        -999,
        -6.5884,
        -5.0326,
        -2.0526,
        -1.0313,
        -0.46489,
        0.060382,
        -0.7508,
        0.53247,
    ]

    return pd.DataFrame(index=index, data=values, columns=["stage_one"])


@pytest.fixture(scope="module")
def skims(settings):
    setup_dirs()
    nw_los = los.Network_LOS()
    nw_los.load_data()
    skim_d = skim_dict(nw_los)

    od_skim_stack_wrapper = skim_d.wrap("origin", "destination")
    do_skim_stack_wrapper = skim_d.wrap("destination", "origin")
    obib_skim_stack_wrapper = skim_d.wrap(tsc.LAST_OB_STOP, tsc.FIRST_IB_STOP)

    skims = [od_skim_stack_wrapper, do_skim_stack_wrapper, obib_skim_stack_wrapper]

    return skims


@pytest.fixture(scope="module")
def locals_dict(skims):
    return {"od_skims": skims[0], "do_skims": skims[1], "obib_skims": skims[2]}


def test_generate_schedule_alternatives(tours):
    windows = tsc.generate_schedule_alternatives(tours)
    assert windows.shape[0] == 296
    assert windows.shape[1] == 4

    output_columns = [
        tsc.SCHEDULE_ID,
        tsc.MAIN_LEG_DURATION,
        tsc.OB_DURATION,
        tsc.IB_DURATION,
    ]

    assert set(output_columns).issubset(windows.columns)


def test_no_stops_patterns(tours):
    no_stops = tours[
        (tours["num_outbound_stops"] == 0) & (tours["num_inbound_stops"] == 0)
    ].copy()
    windows = tsc.no_stops_patterns(no_stops)

    assert windows.shape[0] == 1
    assert windows.shape[1] == 3

    output_columns = [tsc.MAIN_LEG_DURATION, tsc.OB_DURATION, tsc.IB_DURATION]

    assert set(output_columns).issubset(windows.columns)

    pd.testing.assert_series_equal(
        windows[tsc.MAIN_LEG_DURATION],
        no_stops["duration"],
        check_names=False,
        check_dtype=False,
    )
    assert windows[windows[tsc.IB_DURATION] > 0].empty
    assert windows[windows[tsc.OB_DURATION] > 0].empty


def test_one_way_stop_patterns(tours):
    one_way_stops = tours[
        (
            (tours["num_outbound_stops"] > 0).astype(int)
            + (tours["num_inbound_stops"] > 0).astype(int)
        )
        == 1
    ].copy()
    windows = tsc.stop_one_way_only_patterns(one_way_stops)

    assert windows.shape[0] == 58
    assert windows.shape[1] == 3

    output_columns = [tsc.MAIN_LEG_DURATION, tsc.OB_DURATION, tsc.IB_DURATION]

    assert set(output_columns).issubset(windows.columns)

    inbound_options = windows[(windows[tsc.IB_DURATION] > 0)]
    outbound_options = windows[windows[tsc.OB_DURATION] > 0]
    assert np.unique(inbound_options.index).shape[0] == 1
    assert np.unique(outbound_options.index).shape[0] == 1


def test_two_way_stop_patterns(tours):
    two_way_stops = tours[
        (
            (tours["num_outbound_stops"] > 0).astype(int)
            + (tours["num_inbound_stops"] > 0).astype(int)
        )
        == 2
    ].copy()
    windows = tsc.stop_two_way_only_patterns(two_way_stops)

    assert windows.shape[0] == 237
    assert windows.shape[1] == 3

    output_columns = [tsc.MAIN_LEG_DURATION, tsc.OB_DURATION, tsc.IB_DURATION]

    assert set(output_columns).issubset(windows.columns)


def test_run_trip_scheduling_choice(model_spec, tours, skims, locals_dict):
    """
    Test run the model.
    """

    out_tours = tsc.run_trip_scheduling_choice(
        model_spec, tours, skims, locals_dict, 2, None, "PyTest Trip Scheduling"
    )

    assert len(tours) == len(out_tours)
    pd.testing.assert_index_equal(
        tours.sort_index().index, out_tours.sort_index().index
    )

    output_columns = [tsc.MAIN_LEG_DURATION, tsc.OB_DURATION, tsc.IB_DURATION]

    assert set(output_columns).issubset(out_tours.columns)

    assert len(
        out_tours[
            out_tours[output_columns].sum(axis=1) == out_tours[tsc.TOUR_DURATION_COLUMN]
        ]
    ) == len(tours)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import shlex
import subprocess
import traceback

from asv import util
from asv.commands import common_args
from asv.commands.run import Run
from asv.console import log


def _do_build(args):
    env, conf, repo, commit_hash = args
    try:
        with log.set_level(logging.WARN):
            env.install_project(conf, repo, commit_hash)
    except util.ProcessError:
        return (env.name, False)
    return (env.name, True)


def _do_build_multiprocess(args):
    """
    multiprocessing callback to build the project in one particular
    environment.
    """
    try:
        return _do_build(args)
    except BaseException as exc:
        raise util.ParallelFailure(str(exc), exc.__class__, traceback.format_exc())


class Latest(Run):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "latest",
            help="Run a benchmark suite on the HEAD commit",
            description="Run a benchmark suite.",
        )

        common_args.add_bench(parser)
        parser.add_argument(
            "--profile",
            "-p",
            action="store_true",
            help="""In addition to timing, run the benchmarks through
            the `cProfile` profiler and store the results.""",
        )
        common_args.add_parallel(parser)
        common_args.add_show_stderr(parser)
        parser.add_argument(
            "--quick",
            "-q",
            action="store_true",
            help="""Do a "quick" run, where each benchmark function is
            run only once.  This is useful to find basic errors in the
            benchmark functions faster.  The results are unlikely to
            be useful, and thus are not saved.""",
        )
        common_args.add_environment(parser)
        parser.add_argument(
            "--set-commit-hash",
            default=None,
            help="""Set the commit hash to use when recording benchmark
            results. This makes results to be saved also when using an
            existing environment.""",
        )
        common_args.add_launch_method(parser)
        parser.add_argument(
            "--dry-run",
            "-n",
            action="store_true",
            default=None,
            help="""Do not save any results to disk.""",
        )
        common_args.add_machine(parser)
        parser.add_argument(
            "--skip-existing-successful",
            action="store_true",
            help="""Skip running benchmarks that have previous successful
            results""",
        )
        parser.add_argument(
            "--skip-existing-failed",
            action="store_true",
            help="""Skip running benchmarks that have previous failed
            results""",
        )
        parser.add_argument(
            "--skip-existing-commits",
            action="store_true",
            help="""Skip running benchmarks for commits that have existing
            results""",
        )
        parser.add_argument(
            "--skip-existing",
            "-k",
            action="store_true",
            help="""Skip running benchmarks that have previous successful
            or failed results""",
        )
        parser.add_argument(
            "--interleave-processes",
            action="store_true",
            default=False,
            help="""Interleave benchmarks with multiple processes across
            commits. This can avoid measurement biases from commit ordering,
            can take longer.""",
        )
        parser.add_argument(
            "--no-interleave-processes",
            action="store_false",
            dest="interleave_processes",
        )
        parser.add_argument(
            "--no-pull", action="store_true", help="Do not pull the repository"
        )

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(
            conf=conf,
            range_spec="HEAD^!",
            steps=None,
            bench=args.bench,
            attribute=args.attribute,
            parallel=args.parallel,
            show_stderr=args.show_stderr,
            quick=args.quick,
            profile=args.profile,
            env_spec=args.env_spec,
            set_commit_hash=args.set_commit_hash,
            dry_run=args.dry_run,
            machine=args.machine,
            skip_successful=args.skip_existing_successful or args.skip_existing,
            skip_failed=args.skip_existing_failed or args.skip_existing,
            skip_existing_commits=args.skip_existing_commits,
            record_samples=True,
            append_samples=True,
            pull=not args.no_pull,
            interleave_processes=args.interleave_processes,
            launch_method=args.launch_method,
            **kwargs
        )


class Batch(Run):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "batch",
            help="Run a set of benchmark suites based on a batch file. "
            "Simply give the file name, which should be a text file "
            "containing a number of activitysim benchmark commands.",
            description="Run a set of benchmark suites based on a batch file.",
        )

        parser.add_argument(
            "file",
            action="store",
            type=str,
            help="""Set the file name to use for reading multiple commands.""",
        )

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        with open(args.file, "rt") as f:
            for line in f.readlines():
                subprocess.run(["activitysim", "benchmark", *shlex.split(line)])



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, inject, pipeline, simulate, tracing
from activitysim.core.interaction_sample import interaction_sample
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.util import assign_in_place

from .util import estimation, tour_destination

logger = logging.getLogger(__name__)
DUMP = False


@inject.step()
def atwork_subtour_destination(
    tours, persons_merged, network_los, chunk_size, trace_hh_id
):

    trace_label = "atwork_subtour_destination"
    model_settings_file_name = "atwork_subtour_destination.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    future_settings = {
        "SIZE_TERM_SELECTOR": "atwork",
        "SEGMENTS": ["atwork"],
        "ORIG_ZONE_ID": "workplace_zone_id",
    }
    model_settings = config.future_model_settings(
        model_settings_file_name, model_settings, future_settings
    )

    destination_column_name = "destination"
    logsum_column_name = model_settings.get("DEST_CHOICE_LOGSUM_COLUMN_NAME")
    want_logsums = logsum_column_name is not None

    sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    persons_merged = persons_merged.to_frame()

    tours = tours.to_frame()
    subtours = tours[tours.tour_category == "atwork"]

    # - if no atwork subtours
    if subtours.shape[0] == 0:
        tracing.no_results("atwork_subtour_destination")
        return

    estimator = estimation.manager.begin_estimation("atwork_subtour_destination")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        # estimator.write_spec(model_settings, tag='SAMPLE_SPEC')
        estimator.write_spec(model_settings, tag="SPEC")
        estimator.set_alt_id(model_settings["ALT_DEST_COL_NAME"])
        estimator.write_table(
            inject.get_injectable("size_terms"), "size_terms", append=False
        )
        estimator.write_table(
            inject.get_table("land_use").to_frame(), "landuse", append=False
        )
        estimator.write_model_settings(model_settings, model_settings_file_name)

    choices_df, save_sample_df = tour_destination.run_tour_destination(
        subtours,
        persons_merged,
        want_logsums,
        want_sample_table,
        model_settings,
        network_los,
        estimator,
        chunk_size,
        trace_hh_id,
        trace_label,
    )

    if estimator:
        estimator.write_choices(choices_df["choice"])
        choices_df["choice"] = estimator.get_survey_values(
            choices_df["choice"], "tours", "destination"
        )
        estimator.write_override_choices(choices_df["choice"])
        estimator.end_estimation()

    subtours[destination_column_name] = choices_df["choice"]
    assign_in_place(tours, subtours[[destination_column_name]])

    if want_logsums:
        subtours[logsum_column_name] = choices_df["logsum"]
        assign_in_place(tours, subtours[[logsum_column_name]])

    pipeline.replace_table("tours", tours)

    tracing.print_summary(
        destination_column_name, subtours[destination_column_name], describe=True
    )

    if want_sample_table:
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(choices_df)
        # save_sample_df.set_index(model_settings['ALT_DEST_COL_NAME'], append=True, inplace=True)
        pipeline.extend_table(sample_table_name, save_sample_df)

    if trace_hh_id:
        tracing.trace_df(
            tours, label="atwork_subtour_destination", columns=["destination"]
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, simulate, tracing

"""
At this time, these utilities are mostly for transforming the mode choice
spec, which is more complicated than the other specs, into something that
looks like the other specs.
"""

logger = logging.getLogger(__name__)


def mode_choice_simulate(
    choosers,
    spec,
    nest_spec,
    skims,
    locals_d,
    chunk_size,
    mode_column_name,
    logsum_column_name,
    trace_label,
    trace_choice_name,
    trace_column_names=None,
    estimator=None,
):
    """
    common method for  both tour_mode_choice and trip_mode_choice

    Parameters
    ----------
    choosers
    spec
    nest_spec
    skims
    locals_d
    chunk_size
    mode_column_name
    logsum_column_name
    trace_label
    trace_choice_name
    estimator

    Returns
    -------

    """
    want_logsums = logsum_column_name is not None

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=spec,
        nest_spec=nest_spec,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        want_logsums=want_logsums,
        trace_label=trace_label,
        trace_choice_name=trace_choice_name,
        estimator=estimator,
        trace_column_names=trace_column_names,
    )

    # for consistency, always return dataframe, whether or not logsums were requested
    if isinstance(choices, pd.Series):
        choices = choices.to_frame("choice")

    choices.rename(
        columns={"logsum": logsum_column_name, "choice": mode_column_name}, inplace=True
    )

    alts = spec.columns
    choices[mode_column_name] = choices[mode_column_name].map(
        dict(list(zip(list(range(len(alts))), alts)))
    )

    return choices


def run_tour_mode_choice_simulate(
    choosers,
    tour_purpose,
    model_settings,
    mode_column_name,
    logsum_column_name,
    network_los,
    skims,
    constants,
    estimator,
    chunk_size,
    trace_label=None,
    trace_choice_name=None,
):
    """
    This is a utility to run a mode choice model for each segment (usually
    segments are tour/trip purposes).  Pass in the tours/trip that need a mode,
    the Skim object, the spec to evaluate with, and any additional expressions
    you want to use in the evaluation of variables.
    """

    spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients = simulate.get_segment_coefficients(model_settings, tour_purpose)

    spec = simulate.eval_coefficients(spec, coefficients, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients, trace_label)

    locals_dict = {}
    locals_dict.update(constants)
    locals_dict.update(skims)

    # coefficients can appear in expressions
    locals_keys = locals_dict.keys()
    if any([coeff in locals_keys for coeff in coefficients.keys()]):
        logger.warning("coefficients are obscuring locals_dict values")
    locals_dict.update(coefficients)

    assert ("in_period" not in choosers) and ("out_period" not in choosers)
    in_time = skims["in_time_col_name"]
    out_time = skims["out_time_col_name"]
    choosers["in_period"] = network_los.skim_time_period_label(choosers[in_time])
    choosers["out_period"] = network_los.skim_time_period_label(choosers[out_time])

    expressions.annotate_preprocessors(
        choosers, locals_dict, skims, model_settings, trace_label
    )

    trace_column_names = choosers.index.name
    assert trace_column_names == "tour_id"
    if trace_column_names not in choosers:
        choosers[trace_column_names] = choosers.index

    if estimator:
        # write choosers after annotation
        estimator.write_choosers(choosers)

    choices = mode_choice_simulate(
        choosers=choosers,
        spec=spec,
        nest_spec=nest_spec,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        mode_column_name=mode_column_name,
        logsum_column_name=logsum_column_name,
        trace_label=trace_label,
        trace_choice_name=trace_choice_name,
        trace_column_names=trace_column_names,
        estimator=estimator,
    )

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os

import numpy as np
import numpy.testing as npt
import openmatrix as omx
import pandas as pd
import pandas.testing as pdt
import pkg_resources
import pytest
import yaml

from activitysim.core import config, inject, pipeline, random, tracing

# set the max households for all tests (this is to limit memory use on travis)
HOUSEHOLDS_SAMPLE_SIZE = 50
HOUSEHOLDS_SAMPLE_RATE = 0.01  # HOUSEHOLDS_SAMPLE_RATE / 5000 households

# household with mandatory, non mandatory, atwork_subtours, and joint tours
HH_ID = 257341

#  [ 257341 1234246 1402915 1511245 1931827 1931908 2307195 2366390 2408855
# 2518594 2549865  982981 1594365 1057690 1234121 2098971]

# SKIP_FULL_RUN = True
SKIP_FULL_RUN = False


def example_path(dirname):
    resource = os.path.join("examples", "prototype_mtc", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def setup_dirs(ancillary_configs_dir=None, data_dir=None):

    # ancillary_configs_dir is used by run_mp to test multiprocess

    test_pipeline_configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    example_configs_dir = example_path("configs")
    configs_dir = [test_pipeline_configs_dir, example_configs_dir]

    if ancillary_configs_dir is not None:
        configs_dir = [ancillary_configs_dir] + configs_dir

    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)

    if not data_dir:
        data_dir = example_path("data")

    inject.add_injectable("data_dir", data_dir)

    inject.clear_cache()

    tracing.config_logger()

    tracing.delete_output_files("csv")
    tracing.delete_output_files("txt")
    tracing.delete_output_files("yaml")
    tracing.delete_output_files("omx")


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def close_handlers():

    loggers = logging.Logger.manager.loggerDict
    for name in loggers:
        logger = logging.getLogger(name)
        logger.handlers = []
        logger.propagate = True
        logger.setLevel(logging.NOTSET)


def inject_settings(**kwargs):

    settings = config.read_settings_file("settings.yaml", mandatory=True)

    for k in kwargs:
        settings[k] = kwargs[k]

    inject.add_injectable("settings", settings)

    return settings
